import copy
import inspect
import json
from functools import wraps
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    TypedDict,
    Union,
)

from google.genai.types import Content, GenerateContentConfig, Part, PartDict

import sentry_sdk
from sentry_sdk._types import BLOB_DATA_SUBSTITUTE
from sentry_sdk.ai.utils import (
    get_modality_from_mime_type,
    normalize_message_roles,
    set_data_normalized,
    transform_google_content_part,
    truncate_and_annotate_messages,
)
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import (
    has_span_streaming_enabled,
    should_truncate_gen_ai_input,
)
from sentry_sdk.utils import (
    capture_internal_exceptions,
    event_from_exception,
    has_data_collection_enabled,
    safe_serialize,
)

from .consts import GEN_AI_SYSTEM, ORIGIN, TOOL_ATTRIBUTES_MAP

if TYPE_CHECKING:
    from google.genai.types import (
        ContentListUnion,
        ContentUnion,
        ContentUnionDict,
        EmbedContentResponse,
        GenerateContentResponse,
        Model,
        Tool,
    )

    from sentry_sdk._types import TextPart
    from sentry_sdk.tracing import Span

_is_PIL_available = False
try:
    from PIL import Image as PILImage  # type: ignore[import-not-found]

    _is_PIL_available = True
except ImportError:
    pass

# Keys to use when checking to see if a dict provided by the user
# is Part-like (as opposed to a Content or multi-turn conversation entry).
_PART_DICT_KEYS = PartDict.__optional_keys__


class UsageData(TypedDict):
    """Structure for token usage data."""

    input_tokens: int
    input_tokens_cached: int
    output_tokens: int
    output_tokens_reasoning: int
    total_tokens: int


def extract_usage_data(
    response: "Union[GenerateContentResponse, dict[str, Any]]",
) -> "UsageData":
    """Extract usage data from response into a structured format.

    Args:
        response: The GenerateContentResponse object or dictionary containing usage metadata

    Returns:
        UsageData: Dictionary with input_tokens, input_tokens_cached,
                   output_tokens, and output_tokens_reasoning fields
    """
    usage_data = UsageData(
        input_tokens=0,
        input_tokens_cached=0,
        output_tokens=0,
        output_tokens_reasoning=0,
        total_tokens=0,
    )

    # Handle dictionary response (from streaming)
    if isinstance(response, dict):
        usage = response.get("usage_metadata", {})
        if not usage:
            return usage_data

        prompt_tokens = usage.get("prompt_token_count", 0) or 0
        tool_use_prompt_tokens = usage.get("tool_use_prompt_token_count", 0) or 0
        usage_data["input_tokens"] = prompt_tokens + tool_use_prompt_tokens

        cached_tokens = usage.get("cached_content_token_count", 0) or 0
        usage_data["input_tokens_cached"] = cached_tokens

        reasoning_tokens = usage.get("thoughts_token_count", 0) or 0
        usage_data["output_tokens_reasoning"] = reasoning_tokens

        candidates_tokens = usage.get("candidates_token_count", 0) or 0
        # python-genai reports output and reasoning tokens separately
        # reasoning should be sub-category of output tokens
        usage_data["output_tokens"] = candidates_tokens + reasoning_tokens

        total_tokens = usage.get("total_token_count", 0) or 0
        usage_data["total_tokens"] = total_tokens

        return usage_data

    if not hasattr(response, "usage_metadata"):
        return usage_data

    usage = response.usage_metadata

    # Input tokens include both prompt and tool use prompt tokens
    prompt_tokens = getattr(usage, "prompt_token_count", 0) or 0
    tool_use_prompt_tokens = getattr(usage, "tool_use_prompt_token_count", 0) or 0
    usage_data["input_tokens"] = prompt_tokens + tool_use_prompt_tokens

    # Cached input tokens
    cached_tokens = getattr(usage, "cached_content_token_count", 0) or 0
    usage_data["input_tokens_cached"] = cached_tokens

    # Reasoning tokens
    reasoning_tokens = getattr(usage, "thoughts_token_count", 0) or 0
    usage_data["output_tokens_reasoning"] = reasoning_tokens

    # output_tokens = candidates_tokens + reasoning_tokens
    # google-genai reports output and reasoning tokens separately
    candidates_tokens = getattr(usage, "candidates_token_count", 0) or 0
    usage_data["output_tokens"] = candidates_tokens + reasoning_tokens

    total_tokens = getattr(usage, "total_token_count", 0) or 0
    usage_data["total_tokens"] = total_tokens

    return usage_data


def _capture_exception(exc: "Any") -> None:
    """Capture exception with Google GenAI mechanism."""
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "google_genai", "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)


def get_model_name(model: "Union[str, Model]") -> str:
    """Extract model name from model parameter."""
    if isinstance(model, str):
        return model
    # Handle case where model might be an object with a name attribute
    if hasattr(model, "name"):
        return str(model.name)
    return str(model)


def extract_contents_messages(contents: "ContentListUnion") -> "List[Dict[str, Any]]":
    """Extract messages from contents parameter which can have various formats.

    Returns a list of message dictionaries in the format:
    - System: {"role": "system", "content": "string"}
    - User/Assistant: {"role": "user"|"assistant", "content": [{"text": "...", "type": "text"}, ...]}
    """
    if contents is None:
        return []

    messages = []

    # Handle string case
    if isinstance(contents, str):
        return [{"role": "user", "content": contents}]

    # Handle list case
    if isinstance(contents, list):
        if contents and all(_is_part_like(item) for item in contents):
            # All items are parts — merge into a single multi-part user message
            content_parts = []
            for item in contents:
                part = _extract_part_from_item(item)
                if part is not None:
                    content_parts.append(part)

            return [{"role": "user", "content": content_parts}]
        else:
            # Multi-turn conversation or mixed content types
            for item in contents:
                item_messages = extract_contents_messages(item)
                messages.extend(item_messages)
            return messages

    # Handle dictionary case (ContentDict)
    if isinstance(contents, dict):
        role = contents.get("role", "user")
        parts = contents.get("parts")

        if parts:
            content_parts = []
            tool_messages = []

            for part in parts:
                part_result = _extract_part_content(part)
                if part_result is None:
                    continue

                if isinstance(part_result, dict) and part_result.get("role") == "tool":
                    # Tool message - add separately
                    tool_messages.append(part_result)
                else:
                    # Regular content part
                    content_parts.append(part_result)

            # Add main message if we have content parts
            if content_parts:
                # Normalize role: "model" -> "assistant"
                normalized_role = "assistant" if role == "model" else role or "user"
                messages.append({"role": normalized_role, "content": content_parts})

            # Add tool messages
            messages.extend(tool_messages)
        elif "text" in contents:
            messages.append(
                {
                    "role": role,
                    "content": [{"text": contents["text"], "type": "text"}],
                }
            )
        elif "inline_data" in contents:
            # The "data" will always be bytes (or bytes within a string),
            # so if this is present, it's safe to automatically substitute with the placeholder
            messages.append(
                {
                    "inline_data": {
                        "mime_type": contents["inline_data"].get("mime_type", ""),
                        "data": BLOB_DATA_SUBSTITUTE,
                    }
                }
            )

        return messages

    # Handle Content object
    if hasattr(contents, "parts") and contents.parts:
        role = getattr(contents, "role", None) or "user"
        content_parts = []
        tool_messages = []

        for part in contents.parts:
            part_result = _extract_part_content(part)
            if part_result is None:
                continue

            if isinstance(part_result, dict) and part_result.get("role") == "tool":
                tool_messages.append(part_result)
            else:
                content_parts.append(part_result)

        if content_parts:
            normalized_role = "assistant" if role == "model" else role
            messages.append({"role": normalized_role, "content": content_parts})

        messages.extend(tool_messages)
        return messages

    # Handle Part object directly
    part_result = _extract_part_content(contents)
    if part_result:
        if isinstance(part_result, dict) and part_result.get("role") == "tool":
            return [part_result]
        else:
            return [{"role": "user", "content": [part_result]}]

    # Handle PIL.Image.Image
    if _is_PIL_available and isinstance(contents, PILImage.Image):
        blob_part = _extract_pil_image(contents)
        if blob_part:
            return [{"role": "user", "content": [blob_part]}]

    # Handle File object
    if hasattr(contents, "uri") and hasattr(contents, "mime_type"):
        # File object
        file_uri = getattr(contents, "uri", None)
        mime_type = getattr(contents, "mime_type", None)
        # Process if we have file_uri, even if mime_type is missing
        if file_uri is not None:
            # Default to empty string if mime_type is None
            if mime_type is None:
                mime_type = ""

            blob_part = {
                "type": "uri",
                "modality": get_modality_from_mime_type(mime_type),
                "mime_type": mime_type,
                "uri": file_uri,
            }
            return [{"role": "user", "content": [blob_part]}]

    # Handle direct text attribute
    if hasattr(contents, "text") and contents.text:
        return [
            {"role": "user", "content": [{"text": str(contents.text), "type": "text"}]}
        ]

    return []


def _extract_part_content(part: "Any") -> "Optional[dict[str, Any]]":
    """Extract content from a Part object or dict.

    Returns:
        - dict for content part (text/blob) or tool message
        - None if part should be skipped
    """
    if part is None:
        return None

    # Handle dict Part
    if isinstance(part, dict):
        # Check for function_response first (tool message)
        if "function_response" in part:
            return _extract_tool_message_from_part(part)

        if part.get("text"):
            return {"text": part["text"], "type": "text"}

        # Try using Google-specific transform for dict formats (inline_data, file_data)
        result = transform_google_content_part(part)
        if result is not None:
            # For inline_data with bytes data, substitute the content
            if "inline_data" in part:
                # inline_data.data will always be bytes, or a string containing base64-encoded bytes,
                # so can automatically substitute without further checks
                result["content"] = BLOB_DATA_SUBSTITUTE
            return result

        return None

    # Handle Part object
    # Check for function_response (tool message)
    if hasattr(part, "function_response") and part.function_response:
        return _extract_tool_message_from_part(part)

    # Handle text
    if hasattr(part, "text") and part.text:
        return {"text": part.text, "type": "text"}

    # Handle file_data
    if hasattr(part, "file_data") and part.file_data:
        file_data = part.file_data
        file_uri = getattr(file_data, "file_uri", None)
        mime_type = getattr(file_data, "mime_type", None)
        # Process if we have file_uri, even if mime_type is missing (consistent with dict handling)
        if file_uri is not None:
            # Default to empty string if mime_type is None (consistent with transform_google_content_part)
            if mime_type is None:
                mime_type = ""

            return {
                "type": "uri",
                "modality": get_modality_from_mime_type(mime_type),
                "mime_type": mime_type,
                "uri": file_uri,
            }

    # Handle inline_data
    if hasattr(part, "inline_data") and part.inline_data:
        inline_data = part.inline_data
        data = getattr(inline_data, "data", None)
        mime_type = getattr(inline_data, "mime_type", None)
        # Process if we have data, even if mime_type is missing/empty (consistent with dict handling)
        if data is not None:
            # Default to empty string if mime_type is None (consistent with transform_google_content_part)
            if mime_type is None:
                mime_type = ""

            return {
                "type": "blob",
                "modality": get_modality_from_mime_type(mime_type),
                "mime_type": mime_type,
                "content": BLOB_DATA_SUBSTITUTE,
            }

    return None


def _extract_tool_message_from_part(part: "Any") -> "Optional[dict[str, Any]]":
    """Extract tool message from a Part with function_response.

    Returns:
        {"role": "tool", "content": {"toolCallId": "...", "toolName": "...", "output": "..."}}
        or None if not a valid tool message
    """
    function_response = None

    if isinstance(part, dict):
        function_response = part.get("function_response")
    elif hasattr(part, "function_response"):
        function_response = part.function_response

    if not function_response:
        return None

    # Extract fields from function_response
    tool_call_id = None
    tool_name = None
    output = None

    if isinstance(function_response, dict):
        tool_call_id = function_response.get("id")
        tool_name = function_response.get("name")
        response_dict = function_response.get("response", {})
        # Prefer "output" key if present, otherwise use entire response
        output = response_dict.get("output", response_dict)
    else:
        # FunctionResponse object
        tool_call_id = getattr(function_response, "id", None)
        tool_name = getattr(function_response, "name", None)
        response_obj = getattr(function_response, "response", None)
        if response_obj is None:
            response_obj = {}
        if isinstance(response_obj, dict):
            output = response_obj.get("output", response_obj)
        else:
            output = response_obj

    if not tool_name:
        return None

    return {
        "role": "tool",
        "content": {
            "toolCallId": str(tool_call_id) if tool_call_id else None,
            "toolName": str(tool_name),
            "output": safe_serialize(output) if output is not None else None,
        },
    }


def _extract_pil_image(image: "Any") -> "Optional[dict[str, Any]]":
    """Extract blob part from PIL.Image.Image."""
    if not _is_PIL_available or not isinstance(image, PILImage.Image):
        return None

    # Get format, default to JPEG
    format_str = image.format or "JPEG"
    suffix = format_str.lower()
    mime_type = f"image/{suffix}"

    return {
        "type": "blob",
        "modality": get_modality_from_mime_type(mime_type),
        "mime_type": mime_type,
        "content": BLOB_DATA_SUBSTITUTE,
    }


def _is_part_like(item: "Any") -> bool:
    """Check if item is a part-like value (PartUnionDict) rather than a Content/multi-turn entry."""
    if isinstance(item, (str, Part)):
        return True
    if isinstance(item, (list, Content)):
        return False
    if isinstance(item, dict):
        if "role" in item or "parts" in item:
            return False
        # Part objects that came in as plain dicts
        return bool(_PART_DICT_KEYS & item.keys())
    # File objects
    if hasattr(item, "uri"):
        return True
    # PIL.Image
    if _is_PIL_available and isinstance(item, PILImage.Image):
        return True
    return False


def _extract_part_from_item(item: "Any") -> "Optional[dict[str, Any]]":
    """Convert a single part-like item to a content part dict."""
    if isinstance(item, str):
        return {"text": item, "type": "text"}

    # Handle bare inline_data dicts directly to preserve the raw format
    if isinstance(item, dict) and "inline_data" in item:
        return {
            "inline_data": {
                "mime_type": item["inline_data"].get("mime_type", ""),
                "data": BLOB_DATA_SUBSTITUTE,
            }
        }

    # For other dicts and Part objects, use existing _extract_part_content
    result = _extract_part_content(item)
    if result is not None:
        return result

    # PIL.Image
    if _is_PIL_available and isinstance(item, PILImage.Image):
        return _extract_pil_image(item)

    # File objects
    if hasattr(item, "uri") and hasattr(item, "mime_type"):
        file_uri = getattr(item, "uri", None)
        mime_type = getattr(item, "mime_type", None) or ""
        if file_uri is not None:
            return {
                "type": "uri",
                "modality": get_modality_from_mime_type(mime_type),
                "mime_type": mime_type,
                "uri": file_uri,
            }

    return None


def extract_contents_text(contents: "ContentListUnion") -> "Optional[str]":
    """Extract text from contents parameter which can have various formats.

    This is a compatibility function that extracts text from messages.
    For new code, use extract_contents_messages instead.
    """
    messages = extract_contents_messages(contents)
    if not messages:
        return None

    texts = []
    for message in messages:
        content = message.get("content")
        if isinstance(content, str):
            texts.append(content)
        elif isinstance(content, list):
            for part in content:
                if isinstance(part, dict) and part.get("type") == "text":
                    texts.append(part.get("text", ""))

    return " ".join(texts) if texts else None


def _format_tools_for_span(
    tools: "Iterable[Tool | Callable[..., Any]]",
) -> "Optional[List[dict[str, Any]]]":
    """Format tools parameter for span data."""
    formatted_tools = []
    for tool in tools:
        if callable(tool):
            # Handle callable functions passed directly
            formatted_tools.append(
                {
                    "name": getattr(tool, "__name__", "unknown"),
                    "description": getattr(tool, "__doc__", None),
                }
            )
        elif (
            hasattr(tool, "function_declarations")
            and tool.function_declarations is not None
        ):
            # Tool object with function declarations
            for func_decl in tool.function_declarations:
                formatted_tools.append(
                    {
                        "name": getattr(func_decl, "name", None),
                        "description": getattr(func_decl, "description", None),
                    }
                )
        else:
            # Check for predefined tool attributes - each of these tools
            # is an attribute of the tool object, by default set to None
            for attr_name, description in TOOL_ATTRIBUTES_MAP.items():
                if getattr(tool, attr_name, None):
                    formatted_tools.append(
                        {
                            "name": attr_name,
                            "description": description,
                        }
                    )
                    break

    return formatted_tools if formatted_tools else None


def extract_tool_calls(
    response: "GenerateContentResponse",
) -> "Optional[List[dict[str, Any]]]":
    """Extract tool/function calls from response candidates and automatic function calling history."""

    tool_calls = []

    # Extract from candidates, sometimes tool calls are nested under the content.parts object
    if getattr(response, "candidates", []):
        for candidate in response.candidates:
            if not hasattr(candidate, "content") or not getattr(
                candidate.content, "parts", []
            ):
                continue

            for part in candidate.content.parts:
                if getattr(part, "function_call", None):
                    function_call = part.function_call
                    tool_call = {
                        "name": getattr(function_call, "name", None),
                        "type": "function_call",
                    }

                    # Extract arguments if available
                    if getattr(function_call, "args", None):
                        tool_call["arguments"] = safe_serialize(function_call.args)

                    tool_calls.append(tool_call)

    # Extract from automatic_function_calling_history
    # This is the history of tool calls made by the model
    if getattr(response, "automatic_function_calling_history", None):
        for content in response.automatic_function_calling_history:
            if not getattr(content, "parts", None):
                continue

            for part in getattr(content, "parts", []):
                if getattr(part, "function_call", None):
                    function_call = part.function_call
                    tool_call = {
                        "name": getattr(function_call, "name", None),
                        "type": "function_call",
                    }

                    # Extract arguments if available
                    if hasattr(function_call, "args"):
                        tool_call["arguments"] = safe_serialize(function_call.args)

                    tool_calls.append(tool_call)

    return tool_calls if tool_calls else None


def _capture_tool_input(
    args: "tuple[Any, ...]", kwargs: "dict[str, Any]", tool: "Tool"
) -> "dict[str, Any]":
    """Capture tool input from args and kwargs."""
    tool_input = kwargs.copy() if kwargs else {}

    # If we have positional args, try to map them to the function signature
    if args:
        try:
            sig = inspect.signature(tool)
            param_names = list(sig.parameters.keys())
            for i, arg in enumerate(args):
                if i < len(param_names):
                    tool_input[param_names[i]] = arg
        except Exception:
            # Fallback if we can't get the signature
            tool_input["args"] = args

    return tool_input


def _create_tool_span(
    tool_name: str, tool_doc: "Optional[str]"
) -> "Union[Span, StreamedSpan]":
    """Create a span for tool execution."""
    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"execute_tool {tool_name}",
            attributes={
                "sentry.op": OP.GEN_AI_EXECUTE_TOOL,
                "sentry.origin": ORIGIN,
                SPANDATA.GEN_AI_TOOL_NAME: tool_name,
            },
        )
        if tool_doc:
            span.set_attribute(SPANDATA.GEN_AI_TOOL_DESCRIPTION, tool_doc)
        return span

    span = sentry_sdk.start_span(
        op=OP.GEN_AI_EXECUTE_TOOL,
        name=f"execute_tool {tool_name}",
        origin=ORIGIN,
    )
    span.set_data(SPANDATA.GEN_AI_TOOL_NAME, tool_name)
    if tool_doc:
        span.set_data(SPANDATA.GEN_AI_TOOL_DESCRIPTION, tool_doc)
    return span


def wrapped_tool(tool: "Tool | Callable[..., Any]") -> "Tool | Callable[..., Any]":
    """Wrap a tool to emit execute_tool spans when called."""
    if not callable(tool):
        # Not a callable function, return as-is (predefined tools)
        return tool

    tool_name = getattr(tool, "__name__", "unknown")
    tool_doc = tool.__doc__

    if inspect.iscoroutinefunction(tool):
        # Async function
        @wraps(tool)
        async def async_wrapped(*args: "Any", **kwargs: "Any") -> "Any":
            with _create_tool_span(tool_name, tool_doc) as span:
                set_on_span = (
                    span.set_attribute
                    if isinstance(span, StreamedSpan)
                    else span.set_data
                )
                # Capture tool input
                tool_input = _capture_tool_input(args, kwargs, tool)
                with capture_internal_exceptions():
                    set_on_span(SPANDATA.GEN_AI_TOOL_INPUT, safe_serialize(tool_input))

                try:
                    result = await tool(*args, **kwargs)

                    # Capture tool output
                    with capture_internal_exceptions():
                        set_on_span(SPANDATA.GEN_AI_TOOL_OUTPUT, safe_serialize(result))

                    return result
                except Exception as exc:
                    _capture_exception(exc)
                    raise

        return async_wrapped
    else:
        # Sync function
        @wraps(tool)
        def sync_wrapped(*args: "Any", **kwargs: "Any") -> "Any":
            with _create_tool_span(tool_name, tool_doc) as span:
                set_on_span = (
                    span.set_attribute
                    if isinstance(span, StreamedSpan)
                    else span.set_data
                )
                # Capture tool input
                tool_input = _capture_tool_input(args, kwargs, tool)
                with capture_internal_exceptions():
                    set_on_span(SPANDATA.GEN_AI_TOOL_INPUT, safe_serialize(tool_input))

                try:
                    result = tool(*args, **kwargs)

                    # Capture tool output
                    with capture_internal_exceptions():
                        set_on_span(SPANDATA.GEN_AI_TOOL_OUTPUT, safe_serialize(result))

                    return result
                except Exception as exc:
                    _capture_exception(exc)
                    raise

        return sync_wrapped


def wrapped_config_with_tools(
    config: "GenerateContentConfig",
) -> "GenerateContentConfig":
    """Wrap tools in config to emit execute_tool spans. Tools are sometimes passed directly as
    callable functions as a part of the config object."""

    if not config or not getattr(config, "tools", None):
        return config

    result = copy.copy(config)
    result.tools = [wrapped_tool(tool) for tool in config.tools]

    return result


def _extract_response_text(
    response: "GenerateContentResponse",
) -> "Optional[List[str]]":
    """Extract text from response candidates."""

    if not response or not getattr(response, "candidates", []):
        return None

    texts = []
    for candidate in response.candidates:
        if not hasattr(candidate, "content") or not hasattr(candidate.content, "parts"):
            continue

        if candidate.content is None or candidate.content.parts is None:
            continue

        for part in candidate.content.parts:
            if getattr(part, "text", None):
                texts.append(part.text)

    return texts if texts else None


def extract_finish_reasons(
    response: "GenerateContentResponse",
) -> "Optional[List[str]]":
    """Extract finish reasons from response candidates."""
    if not response or not getattr(response, "candidates", []):
        return None

    finish_reasons = []
    for candidate in response.candidates:
        if getattr(candidate, "finish_reason", None):
            # Convert enum value to string if necessary
            reason = str(candidate.finish_reason)
            # Remove enum prefix if present (e.g., "FinishReason.STOP" -> "STOP")
            if "." in reason:
                reason = reason.split(".")[-1]
            finish_reasons.append(reason)

    return finish_reasons if finish_reasons else None


def _transform_system_instruction_one_level(
    system_instructions: "Union[ContentUnionDict, ContentUnion]",
    can_be_content: bool,
) -> "list[TextPart]":
    text_parts: "list[TextPart]" = []

    if isinstance(system_instructions, str):
        return [{"type": "text", "content": system_instructions}]

    if isinstance(system_instructions, Part) and system_instructions.text:
        return [{"type": "text", "content": system_instructions.text}]

    if can_be_content and isinstance(system_instructions, Content):
        if isinstance(system_instructions.parts, list):
            for part in system_instructions.parts:
                if isinstance(part.text, str):
                    text_parts.append({"type": "text", "content": part.text})
        return text_parts

    if isinstance(system_instructions, dict) and system_instructions.get("text"):
        return [{"type": "text", "content": system_instructions["text"]}]

    elif can_be_content and isinstance(system_instructions, dict):
        parts = system_instructions.get("parts", [])
        for part in parts:
            if isinstance(part, Part) and isinstance(part.text, str):
                text_parts.append({"type": "text", "content": part.text})
            elif isinstance(part, dict) and isinstance(part.get("text"), str):
                text_parts.append({"type": "text", "content": part["text"]})
        return text_parts

    return text_parts


def _transform_system_instructions(
    system_instructions: "Union[ContentUnionDict, ContentUnion]",
) -> "list[TextPart]":
    text_parts: "list[TextPart]" = []

    if isinstance(system_instructions, list):
        text_parts = list(
            chain.from_iterable(
                _transform_system_instruction_one_level(
                    instructions, can_be_content=False
                )
                for instructions in system_instructions
            )
        )

        return text_parts

    return _transform_system_instruction_one_level(
        system_instructions, can_be_content=True
    )


def set_span_data_for_request(
    span: "Union[Span, StreamedSpan]",
    integration: "Any",
    model: str,
    contents: "ContentListUnion",
    kwargs: "dict[str, Any]",
) -> None:
    """Set span data for the request."""
    client = sentry_sdk.get_client()
    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    set_on_span(SPANDATA.GEN_AI_SYSTEM, GEN_AI_SYSTEM)
    set_on_span(SPANDATA.GEN_AI_REQUEST_MODEL, model)

    if kwargs.get("stream", False):
        set_on_span(SPANDATA.GEN_AI_RESPONSE_STREAMING, True)

    config: "Optional[GenerateContentConfig]" = kwargs.get("config")

    # Set tools if available
    if config is not None and hasattr(config, "tools"):
        tools = config.tools
        if tools:
            formatted_tools = _format_tools_for_span(tools)
            if formatted_tools:
                if has_data_collection_enabled(client.options):
                    if client.options["data_collection"]["gen_ai"]["inputs"]:
                        set_data_normalized(
                            span,
                            SPANDATA.GEN_AI_REQUEST_AVAILABLE_TOOLS,
                            formatted_tools,
                            unpack=False,
                        )
                else:
                    # To remove once data collection has been fully rolled out
                    set_data_normalized(
                        span,
                        SPANDATA.GEN_AI_REQUEST_AVAILABLE_TOOLS,
                        formatted_tools,
                        unpack=False,
                    )

    record_inputs = False
    if has_data_collection_enabled(client.options):
        if client.options["data_collection"]["gen_ai"]["inputs"]:
            record_inputs = True
    elif should_send_default_pii() and integration.include_prompts:
        record_inputs = True

    if record_inputs:
        messages = []

        # Add system instruction if present
        system_instructions = None
        if config and hasattr(config, "system_instruction"):
            system_instructions = config.system_instruction
        elif isinstance(config, dict) and "system_instruction" in config:
            system_instructions = config.get("system_instruction")

        if system_instructions is not None:
            set_on_span(
                SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
                json.dumps(_transform_system_instructions(system_instructions)),
            )

        # Extract messages from contents
        contents_messages = extract_contents_messages(contents)
        messages.extend(contents_messages)

        if messages:
            normalized_messages = normalize_message_roles(messages)
            client = sentry_sdk.get_client()
            scope = sentry_sdk.get_current_scope()
            messages_data = (
                truncate_and_annotate_messages(normalized_messages, span, scope)
                if should_truncate_gen_ai_input(client.options)
                else normalized_messages
            )
            if messages_data is not None:
                set_data_normalized(
                    span,
                    SPANDATA.GEN_AI_REQUEST_MESSAGES,
                    messages_data,
                    unpack=False,
                )

    # Extract parameters directly from config (not nested under generation_config)
    for param, span_key in [
        ("temperature", SPANDATA.GEN_AI_REQUEST_TEMPERATURE),
        ("top_p", SPANDATA.GEN_AI_REQUEST_TOP_P),
        ("top_k", SPANDATA.GEN_AI_REQUEST_TOP_K),
        ("max_output_tokens", SPANDATA.GEN_AI_REQUEST_MAX_TOKENS),
        ("presence_penalty", SPANDATA.GEN_AI_REQUEST_PRESENCE_PENALTY),
        ("frequency_penalty", SPANDATA.GEN_AI_REQUEST_FREQUENCY_PENALTY),
        ("seed", SPANDATA.GEN_AI_REQUEST_SEED),
    ]:
        if hasattr(config, param):
            value = getattr(config, param)
            if value is not None:
                set_on_span(span_key, value)


def set_span_data_for_response(
    span: "Union[Span, StreamedSpan]",
    integration: "Any",
    response: "GenerateContentResponse",
) -> None:
    if not response:
        return

    client = sentry_sdk.get_client()
    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )

    finish_reasons = extract_finish_reasons(response)
    if finish_reasons:
        set_data_normalized(
            span, SPANDATA.GEN_AI_RESPONSE_FINISH_REASONS, finish_reasons
        )

    if getattr(response, "response_id", None):
        set_on_span(SPANDATA.GEN_AI_RESPONSE_ID, response.response_id)

    if getattr(response, "model_version", None):
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, response.model_version)

    usage_data = extract_usage_data(response)

    if usage_data["input_tokens"]:
        set_on_span(SPANDATA.GEN_AI_USAGE_INPUT_TOKENS, usage_data["input_tokens"])

    if usage_data["input_tokens_cached"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_INPUT_TOKENS_CACHED,
            usage_data["input_tokens_cached"],
        )

    if usage_data["output_tokens"]:
        set_on_span(SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS, usage_data["output_tokens"])

    if usage_data["output_tokens_reasoning"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS_REASONING,
            usage_data["output_tokens_reasoning"],
        )

    if usage_data["total_tokens"]:
        set_on_span(SPANDATA.GEN_AI_USAGE_TOTAL_TOKENS, usage_data["total_tokens"])

    tool_calls = extract_tool_calls(response)
    if tool_calls:
        if has_data_collection_enabled(client.options):
            if client.options["data_collection"]["gen_ai"]["outputs"]:
                set_on_span(
                    SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS, safe_serialize(tool_calls)
                )
        else:
            # Before data collection was introduced, this was set unconditionally
            set_on_span(SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS, safe_serialize(tool_calls))

    if has_data_collection_enabled(client.options):
        if client.options["data_collection"]["gen_ai"]["outputs"]:
            response_texts = _extract_response_text(response)
            if response_texts:
                set_on_span(
                    SPANDATA.GEN_AI_RESPONSE_TEXT, safe_serialize(response_texts)
                )
    elif should_send_default_pii() and integration.include_prompts:
        # TODO: Delete this block once data collection has been completely rolled out
        response_texts = _extract_response_text(response)
        if response_texts:
            # Format as JSON string array as per documentation
            set_on_span(SPANDATA.GEN_AI_RESPONSE_TEXT, safe_serialize(response_texts))


def prepare_generate_content_args(
    args: "tuple[Any, ...]", kwargs: "dict[str, Any]"
) -> "tuple[Any, Any, str]":
    """Extract and prepare common arguments for generate_content methods."""
    model = args[0] if args else kwargs.get("model", "unknown")
    contents = args[1] if len(args) > 1 else kwargs.get("contents")
    model_name = get_model_name(model)

    config = kwargs.get("config")
    wrapped_config = wrapped_config_with_tools(config)
    if wrapped_config is not config:
        kwargs["config"] = wrapped_config

    return model, contents, model_name


def prepare_embed_content_args(
    args: "tuple[Any, ...]", kwargs: "dict[str, Any]"
) -> "tuple[str, Any]":
    """Extract and prepare common arguments for embed_content methods.

    Returns:
        tuple: (model_name, contents)
    """
    model = kwargs.get("model", "unknown")
    contents = kwargs.get("contents")
    model_name = get_model_name(model)

    return model_name, contents


def set_span_data_for_embed_request(
    span: "Union[Span, StreamedSpan]",
    integration: "Any",
    contents: "Any",
    kwargs: "dict[str, Any]",
) -> None:
    """Set span data for embedding request."""
    client = sentry_sdk.get_client()

    record_inputs = False
    if has_data_collection_enabled(client.options):
        if client.options["data_collection"]["gen_ai"]["inputs"]:
            record_inputs = True
    elif should_send_default_pii() and integration.include_prompts:
        record_inputs = True

    if record_inputs:
        if contents:
            # For embeddings, contents is typically a list of strings/texts
            input_texts = []

            # Handle various content formats
            if isinstance(contents, str):
                input_texts = [contents]
            elif isinstance(contents, list):
                for item in contents:
                    text = extract_contents_text(item)
                    if text:
                        input_texts.append(text)
            else:
                text = extract_contents_text(contents)
                if text:
                    input_texts = [text]

            if input_texts:
                set_data_normalized(
                    span,
                    SPANDATA.GEN_AI_EMBEDDINGS_INPUT,
                    input_texts,
                    unpack=False,
                )


def set_span_data_for_embed_response(
    span: "Union[Span, StreamedSpan]",
    integration: "Any",
    response: "EmbedContentResponse",
) -> None:
    """Set span data for embedding response."""
    if not response:
        return

    # Extract token counts from embeddings statistics (Vertex AI only)
    # Each embedding has its own statistics with token_count
    if hasattr(response, "embeddings") and response.embeddings:
        total_tokens = 0

        for embedding in response.embeddings:
            if hasattr(embedding, "statistics") and embedding.statistics:
                token_count = getattr(embedding.statistics, "token_count", None)
                if token_count is not None:
                    total_tokens += int(token_count)

        # Set token count if we found any
        if total_tokens > 0:
            if isinstance(span, StreamedSpan):
                span.set_attribute(SPANDATA.GEN_AI_USAGE_INPUT_TOKENS, total_tokens)
            else:
                span.set_data(SPANDATA.GEN_AI_USAGE_INPUT_TOKENS, total_tokens)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.ai.utils import (
    get_start_span_function,
    normalize_message_roles,
    set_data_normalized,
    truncate_and_annotate_messages,
)
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import (
    has_span_streaming_enabled,
    should_truncate_gen_ai_input,
)

from ..consts import SPAN_ORIGIN
from ..utils import (
    _set_agent_data,
    _set_available_tools,
    _set_model_data,
    _should_send_inputs,
    _should_send_outputs,
)
from .utils import (
    _serialize_binary_content_item,
    _serialize_image_url_item,
)

if TYPE_CHECKING:
    from typing import Any, Union

try:
    from pydantic_ai.messages import BinaryContent, ImageUrl
except ImportError:
    BinaryContent = None  # type: ignore[misc,assignment]
    ImageUrl = None  # type: ignore[misc,assignment]


def invoke_agent_span(
    user_prompt: "Any",
    agent: "Any",
    model: "Any",
    model_settings: "Any",
    is_streaming: bool = False,
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    """Create a span for invoking the agent."""
    # Determine agent name for span
    name = "agent"
    if agent and getattr(agent, "name", None):
        name = agent.name

    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"invoke_agent {name}",
            attributes={
                "sentry.op": OP.GEN_AI_INVOKE_AGENT,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "invoke_agent",
            },
        )
    else:
        span = get_start_span_function()(
            op=OP.GEN_AI_INVOKE_AGENT,
            name=f"invoke_agent {name}",
            origin=SPAN_ORIGIN,
        )

        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "invoke_agent")

    _set_agent_data(span, agent)
    _set_model_data(span, model, model_settings)
    _set_available_tools(span, agent)

    # Add user prompt and system prompts if available and prompts are enabled
    if _should_send_inputs():
        messages = []

        # Add system prompts (both instructions and system_prompt)
        system_texts = []

        if agent:
            # Check for system_prompt
            system_prompts = getattr(agent, "_system_prompts", None) or []
            for prompt in system_prompts:
                if isinstance(prompt, str):
                    system_texts.append(prompt)

            # Check for instructions (stored in _instructions)
            instructions = getattr(agent, "_instructions", None)
            if instructions:
                if isinstance(instructions, str):
                    system_texts.append(instructions)
                elif isinstance(instructions, (list, tuple)):
                    for instr in instructions:
                        if isinstance(instr, str):
                            system_texts.append(instr)
                        elif callable(instr):
                            # Skip dynamic/callable instructions
                            pass

        # Add all system texts as system messages
        for system_text in system_texts:
            messages.append(
                {
                    "content": [{"text": system_text, "type": "text"}],
                    "role": "system",
                }
            )

        # Add user prompt
        if user_prompt:
            if isinstance(user_prompt, str):
                messages.append(
                    {
                        "content": [{"text": user_prompt, "type": "text"}],
                        "role": "user",
                    }
                )
            elif isinstance(user_prompt, list):
                # Handle list of user content
                content = []
                for item in user_prompt:
                    if isinstance(item, str):
                        content.append({"text": item, "type": "text"})
                    elif ImageUrl is not None and isinstance(item, ImageUrl):
                        content.append(_serialize_image_url_item(item))
                    elif BinaryContent is not None and isinstance(item, BinaryContent):
                        content.append(_serialize_binary_content_item(item))
                if content:
                    messages.append(
                        {
                            "content": content,
                            "role": "user",
                        }
                    )

        if messages:
            normalized_messages = normalize_message_roles(messages)
            client = sentry_sdk.get_client()
            scope = sentry_sdk.get_current_scope()
            messages_data = (
                truncate_and_annotate_messages(normalized_messages, span, scope)
                if should_truncate_gen_ai_input(client.options)
                else normalized_messages
            )
            set_data_normalized(
                span, SPANDATA.GEN_AI_REQUEST_MESSAGES, messages_data, unpack=False
            )

    return span


def update_invoke_agent_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    result: "Any",
) -> None:
    """Update and close the invoke agent span."""
    if not span or not result:
        return

    # Extract output from result
    output = getattr(result, "output", None)

    # Set response text if prompts are enabled
    if _should_send_outputs() and output:
        set_data_normalized(
            span, SPANDATA.GEN_AI_RESPONSE_TEXT, str(output), unpack=False
        )

    # Set model name from response if available
    if hasattr(result, "response"):
        try:
            response = result.response
            if hasattr(response, "model_name") and response.model_name:
                if isinstance(span, StreamedSpan):
                    span.set_attribute(
                        SPANDATA.GEN_AI_RESPONSE_MODEL, response.model_name
                    )
                else:
                    span.set_data(SPANDATA.GEN_AI_RESPONSE_MODEL, response.model_name)
        except Exception:
            # If response access fails, continue without setting model name
            pass



"""
IMPORTANT: The contents of this file are part of a proof of concept and as such
are experimental and not suitable for production use. They may be changed or
removed at any time without prior notice.
"""

import warnings
from typing import TYPE_CHECKING

from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations.opentelemetry.propagator import SentryPropagator
from sentry_sdk.integrations.opentelemetry.span_processor import SentrySpanProcessor
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import logger

if TYPE_CHECKING:
    from typing import Any, Dict, Optional

try:
    from opentelemetry import trace
    from opentelemetry.propagate import set_global_textmap
    from opentelemetry.sdk.trace import TracerProvider
except ImportError:
    raise DidNotEnable("opentelemetry not installed")

try:
    from opentelemetry.instrumentation.django import (  # type: ignore[import-not-found]
        DjangoInstrumentor,
    )
except ImportError:
    DjangoInstrumentor = None


CONFIGURABLE_INSTRUMENTATIONS = {
    DjangoInstrumentor: {"is_sql_commentor_enabled": True},
}


class OpenTelemetryIntegration(Integration):
    identifier = "opentelemetry"

    @staticmethod
    def setup_once() -> None:
        pass

    def setup_once_with_options(
        self, options: "Optional[Dict[str, Any]]" = None
    ) -> None:
        if has_span_streaming_enabled(options):
            logger.warning(
                "[OTel] OpenTelemetryIntegration is not compatible with span streaming "
                "(trace_lifecycle='stream') and will be disabled."
            )
            return

        warnings.warn(
            "OpenTelemetryIntegration is deprecated. "
            "Please use OTLPIntegration instead: "
            "https://docs.sentry.io/platforms/python/integrations/otlp/",
            DeprecationWarning,
            stacklevel=2,
        )

        _setup_sentry_tracing()
        # _setup_instrumentors()

        logger.debug("[OTel] Finished setting up OpenTelemetry integration")


def _setup_sentry_tracing() -> None:
    provider = TracerProvider()
    provider.add_span_processor(SentrySpanProcessor())
    trace.set_tracer_provider(provider)
    set_global_textmap(SentryPropagator())


def _setup_instrumentors() -> None:
    for instrumentor, kwargs in CONFIGURABLE_INSTRUMENTATIONS.items():
        if instrumentor is None:
            continue
        instrumentor().instrument(**kwargs)



import asyncio
import gzip
import io
import json
import logging
import os
import socket
import ssl
import time
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from urllib.request import getproxies

try:
    import brotli  # type: ignore
except ImportError:
    brotli = None

try:
    import httpcore
except ImportError:
    httpcore = None  # type: ignore[assignment,unused-ignore]

try:
    import h2  # noqa: F401

    HTTP2_ENABLED = httpcore is not None
except ImportError:
    HTTP2_ENABLED = False

try:
    import anyio  # noqa: F401

    ASYNC_TRANSPORT_AVAILABLE = httpcore is not None
except ImportError:
    ASYNC_TRANSPORT_AVAILABLE = False

from typing import TYPE_CHECKING, Dict, List, cast

import certifi
import urllib3

import sentry_sdk
from sentry_sdk.consts import EndpointType
from sentry_sdk.envelope import Envelope, Item, PayloadRef
from sentry_sdk.utils import (
    Dsn,
    capture_internal_exceptions,
    logger,
    mark_sentry_task_internal,
)
from sentry_sdk.worker import AsyncWorker, BackgroundWorker, Worker

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        DefaultDict,
        Iterable,
        Mapping,
        Optional,
        Self,
        Tuple,
        Type,
        Union,
    )

    from urllib3.poolmanager import PoolManager, ProxyManager

    from sentry_sdk._types import Event, EventDataCategory

KEEP_ALIVE_SOCKET_OPTIONS = []
for option in [
    (socket.SOL_SOCKET, lambda: getattr(socket, "SO_KEEPALIVE"), 1),  # noqa: B009
    (socket.SOL_TCP, lambda: getattr(socket, "TCP_KEEPIDLE"), 45),  # noqa: B009
    (socket.SOL_TCP, lambda: getattr(socket, "TCP_KEEPINTVL"), 10),  # noqa: B009
    (socket.SOL_TCP, lambda: getattr(socket, "TCP_KEEPCNT"), 6),  # noqa: B009
]:
    try:
        KEEP_ALIVE_SOCKET_OPTIONS.append((option[0], option[1](), option[2]))
    except AttributeError:
        # a specific option might not be available on specific systems,
        # e.g. TCP_KEEPIDLE doesn't exist on macOS
        pass


def _get_httpcore_header_value(response: "Any", header: str) -> "Optional[str]":
    """Case-insensitive header lookup for httpcore-style responses."""
    header_lower = header.lower()
    return next(
        (
            val.decode("ascii")
            for key, val in response.headers
            if key.decode("ascii").lower() == header_lower
        ),
        None,
    )


class Transport(ABC):
    """Baseclass for all transports.

    A transport is used to send an event to sentry.
    """

    parsed_dsn: "Optional[Dsn]" = None

    def __init__(self: "Self", options: "Optional[Dict[str, Any]]" = None) -> None:
        self.options = options
        if options and options["dsn"]:
            self.parsed_dsn = Dsn(options["dsn"], options.get("org_id"))
        else:
            self.parsed_dsn = None

    def capture_event(self: "Self", event: "Event") -> None:
        """
        DEPRECATED: Please use capture_envelope instead.

        This gets invoked with the event dictionary when an event should
        be sent to sentry.
        """

        warnings.warn(
            "capture_event is deprecated, please use capture_envelope instead!",
            DeprecationWarning,
            stacklevel=2,
        )

        envelope = Envelope()
        envelope.add_event(event)
        self.capture_envelope(envelope)

    @abstractmethod
    def capture_envelope(self: "Self", envelope: "Envelope") -> None:
        """
        Send an envelope to Sentry.

        Envelopes are a data container format that can hold any type of data
        submitted to Sentry. We use it to send all event data (including errors,
        transactions, crons check-ins, etc.) to Sentry.
        """
        pass

    def flush(
        self: "Self",
        timeout: float,
        callback: "Optional[Any]" = None,
    ) -> None:
        """
        Wait `timeout` seconds for the current events to be sent out.

        The default implementation is a no-op, since this method may only be relevant to some transports.
        Subclasses should override this method if necessary.
        """
        return None

    def kill(self: "Self") -> None:
        """
        Forcefully kills the transport.

        The default implementation is a no-op, since this method may only be relevant to some transports.
        Subclasses should override this method if necessary.
        """
        return None

    def record_lost_event(
        self,
        reason: str,
        data_category: "Optional[EventDataCategory]" = None,
        item: "Optional[Item]" = None,
        *,
        quantity: int = 1,
    ) -> None:
        """This increments a counter for event loss by reason and
        data category by the given positive-int quantity (default 1).

        If an item is provided, the data category and quantity are
        extracted from the item, and the values passed for
        data_category and quantity are ignored.

        When recording a lost transaction via data_category="transaction",
        the calling code should also record the lost spans via this method.
        When recording lost spans, `quantity` should be set to the number
        of contained spans, plus one for the transaction itself. When
        passing an Item containing a transaction via the `item` parameter,
        this method automatically records the lost spans.
        """
        return None

    def is_healthy(self: "Self") -> bool:
        return True


def _parse_rate_limits(
    header: str, now: "Optional[datetime]" = None
) -> "Iterable[Tuple[Optional[EventDataCategory], datetime]]":
    if now is None:
        now = datetime.now(timezone.utc)

    for limit in header.split(","):
        try:
            parameters = limit.strip().split(":")
            retry_after_val, categories = parameters[:2]

            retry_after = now + timedelta(seconds=int(retry_after_val))
            for category in categories and categories.split(";") or (None,):
                yield category, retry_after  # type: ignore
        except (LookupError, ValueError):
            continue


class HttpTransportCore(Transport):
    """Shared base class for sync and async transports."""

    TIMEOUT = 30  # seconds

    def __init__(self: "Self", options: "Dict[str, Any]") -> None:
        from sentry_sdk.consts import VERSION

        Transport.__init__(self, options)
        assert self.parsed_dsn is not None
        self.options: "Dict[str, Any]" = options
        self._worker = self._create_worker(options)
        self._auth = self.parsed_dsn.to_auth("sentry.python/%s" % VERSION)
        self._disabled_until: "Dict[Optional[EventDataCategory], datetime]" = {}
        # We only use this Retry() class for the `get_retry_after` method it exposes
        self._retry = urllib3.util.Retry()
        self._discarded_events: "DefaultDict[Tuple[EventDataCategory, str], int]" = (
            defaultdict(int)
        )
        self._last_client_report_sent = time.time()

        self._pool = self._make_pool()

        # Backwards compatibility for deprecated `self.hub_class` attribute
        self._hub_cls = sentry_sdk.Hub

        experiments = options.get("_experiments", {})
        compression_level = experiments.get(
            "transport_compression_level",
            experiments.get("transport_zlib_compression_level"),
        )
        compression_algo = experiments.get(
            "transport_compression_algo",
            (
                "gzip"
                # if only compression level is set, assume gzip for backwards compatibility
                # if we don't have brotli available, fallback to gzip
                if compression_level is not None or brotli is None
                else "br"
            ),
        )

        if compression_algo == "br" and brotli is None:
            logger.warning(
                "You asked for brotli compression without the Brotli module, falling back to gzip -9"
            )
            compression_algo = "gzip"
            compression_level = None

        if compression_algo not in ("br", "gzip"):
            logger.warning(
                "Unknown compression algo %s, disabling compression", compression_algo
            )
            self._compression_level = 0
            self._compression_algo = None
        else:
            self._compression_algo = compression_algo

        if compression_level is not None:
            self._compression_level = compression_level
        elif self._compression_algo == "gzip":
            self._compression_level = 9
        elif self._compression_algo == "br":
            self._compression_level = 4

    def _create_worker(self: "Self", options: "Dict[str, Any]") -> "Worker":
        raise NotImplementedError()

    def record_lost_event(
        self,
        reason: str,
        data_category: "Optional[EventDataCategory]" = None,
        item: "Optional[Item]" = None,
        *,
        quantity: int = 1,
    ) -> None:
        if not self.options["send_client_reports"]:
            return

        if item is not None:
            data_category = item.data_category
            quantity = 1  # If an item is provided, we always count it as 1 (except for attachments, handled below).

            if data_category == "transaction":
                # Also record the lost spans
                event = item.get_transaction_event() or {}

                # +1 for the transaction itself
                span_count = (
                    len(cast(List[Dict[str, object]], event.get("spans") or [])) + 1
                )
                self.record_lost_event(reason, "span", quantity=span_count)

            elif data_category == "log_item" and item:
                # Also record size of lost logs in bytes
                bytes_size = len(item.get_bytes())
                self.record_lost_event(reason, "log_byte", quantity=bytes_size)

            elif data_category == "attachment":
                # quantity of 0 is actually 1 as we do not want to count
                # empty attachments as actually empty.
                quantity = len(item.get_bytes()) or 1

        elif data_category is None:
            raise TypeError("data category not provided")

        self._discarded_events[data_category, reason] += quantity

    def _get_header_value(
        self: "Self", response: "Any", header: str
    ) -> "Optional[str]":
        return response.headers.get(header)

    def _update_rate_limits(
        self: "Self", response: "Union[urllib3.BaseHTTPResponse, httpcore.Response]"
    ) -> None:
        # new sentries with more rate limit insights.  We honor this header
        # no matter of the status code to update our internal rate limits.
        header = self._get_header_value(response, "x-sentry-rate-limits")
        if header:
            logger.warning("Rate-limited via x-sentry-rate-limits")
            self._disabled_until.update(_parse_rate_limits(header))

        # old sentries only communicate global rate limit hits via the
        # retry-after header on 429.  This header can also be emitted on new
        # sentries if a proxy in front wants to globally slow things down.
        elif response.status == 429:
            logger.warning("Rate-limited via 429")
            retry_after_value = self._get_header_value(response, "Retry-After")
            retry_after = (
                self._retry.parse_retry_after(retry_after_value)
                if retry_after_value is not None
                else None
            ) or 60
            self._disabled_until[None] = datetime.now(timezone.utc) + timedelta(
                seconds=retry_after
            )

    def _handle_request_error(
        self: "Self",
        envelope: "Optional[Envelope]",
        loss_reason: str = "network",
        record_reason: str = "network_error",
    ) -> None:
        def record_loss(reason: str) -> None:
            if envelope is None:
                self.record_lost_event(reason, data_category="error")
            else:
                for item in envelope.items:
                    self.record_lost_event(reason, item=item)

        self.on_dropped_event(loss_reason)
        record_loss(record_reason)

    def _handle_response(
        self: "Self",
        response: "Union[urllib3.BaseHTTPResponse, httpcore.Response]",
        envelope: "Optional[Envelope]",
    ) -> None:
        self._update_rate_limits(response)

        if response.status == 413:
            size_exceeded_message = (
                "HTTP 413: Event dropped due to exceeded envelope size limit"
            )
            response_message = getattr(
                response, "data", getattr(response, "content", None)
            )
            if response_message is not None:
                size_exceeded_message += f" (body: {response_message})"

            logger.error(size_exceeded_message)
            self._handle_request_error(
                envelope=envelope, loss_reason="status_413", record_reason="send_error"
            )

        elif response.status == 429:
            # if we hit a 429.  Something was rate limited but we already
            # acted on this in `self._update_rate_limits`.  Note that we
            # do not want to record event loss here as we will have recorded
            # an outcome in relay already.
            self.on_dropped_event("status_429")
            pass

        elif response.status >= 300 or response.status < 200:
            logger.error(
                "Unexpected status code: %s (body: %s)",
                response.status,
                getattr(response, "data", getattr(response, "content", None)),
            )
            self._handle_request_error(
                envelope=envelope, loss_reason="status_{}".format(response.status)
            )

    def _update_headers(
        self: "Self",
        headers: "Dict[str, str]",
    ) -> None:
        headers.update(
            {
                "User-Agent": str(self._auth.client),
                "X-Sentry-Auth": str(self._auth.to_header()),
            }
        )

    def on_dropped_event(self: "Self", _reason: str) -> None:
        return None

    def _fetch_pending_client_report(
        self: "Self", force: bool = False, interval: int = 60
    ) -> "Optional[Item]":
        if not self.options["send_client_reports"]:
            return None

        if not (force or self._last_client_report_sent < time.time() - interval):
            return None

        discarded_events = self._discarded_events
        self._discarded_events = defaultdict(int)
        self._last_client_report_sent = time.time()

        if not discarded_events:
            return None

        return Item(
            PayloadRef(
                json={
                    "timestamp": time.time(),
                    "discarded_events": [
                        {"reason": reason, "category": category, "quantity": quantity}
                        for (
                            (category, reason),
                            quantity,
                        ) in discarded_events.items()
                    ],
                }
            ),
            type="client_report",
        )

    def _check_disabled(self, category: str) -> bool:
        def _disabled(bucket: "Any") -> bool:
            ts = self._disabled_until.get(bucket)
            return ts is not None and ts > datetime.now(timezone.utc)

        return _disabled(category) or _disabled(None)

    def _is_rate_limited(self: "Self") -> bool:
        return any(
            ts > datetime.now(timezone.utc) for ts in self._disabled_until.values()
        )

    def _is_worker_full(self: "Self") -> bool:
        return self._worker.full()

    def is_healthy(self: "Self") -> bool:
        return not (self._is_worker_full() or self._is_rate_limited())

    def _prepare_envelope(
        self: "Self", envelope: "Envelope"
    ) -> "Optional[Tuple[Envelope, io.BytesIO, Dict[str, str]]]":
        # remove all items from the envelope which are over quota
        new_items = []
        for item in envelope.items:
            if self._check_disabled(item.data_category):
                if item.data_category in ("transaction", "error", "default", "statsd"):
                    self.on_dropped_event("self_rate_limits")
                self.record_lost_event("ratelimit_backoff", item=item)
            else:
                new_items.append(item)

        # Since we're modifying the envelope here make a copy so that others
        # that hold references do not see their envelope modified.
        envelope = Envelope(headers=envelope.headers, items=new_items)

        if not envelope.items:
            return None

        # since we're already in the business of sending out an envelope here
        # check if we have one pending for the stats session envelopes so we
        # can attach it to this enveloped scheduled for sending.  This will
        # currently typically attach the client report to the most recent
        # session update.
        client_report_item = self._fetch_pending_client_report(interval=30)
        if client_report_item is not None:
            envelope.items.append(client_report_item)

        content_encoding, body = self._serialize_envelope(envelope)

        assert self.parsed_dsn is not None
        logger.debug(
            "Sending envelope [%s] project:%s host:%s",
            envelope.description,
            self.parsed_dsn.project_id,
            self.parsed_dsn.host,
        )

        headers: "Dict[str, str]" = {
            "Content-Type": "application/x-sentry-envelope",
        }
        if content_encoding:
            headers["Content-Encoding"] = content_encoding

        return envelope, body, headers

    def _serialize_envelope(
        self: "Self", envelope: "Envelope"
    ) -> "tuple[Optional[str], io.BytesIO]":
        content_encoding = None
        body = io.BytesIO()
        if self._compression_level == 0 or self._compression_algo is None:
            envelope.serialize_into(body)
        else:
            content_encoding = self._compression_algo
            if self._compression_algo == "br" and brotli is not None:
                body.write(
                    brotli.compress(
                        envelope.serialize(), quality=self._compression_level
                    )
                )
            else:  # assume gzip as we sanitize the algo value in init
                with gzip.GzipFile(
                    fileobj=body, mode="w", compresslevel=self._compression_level
                ) as f:
                    envelope.serialize_into(f)

        return content_encoding, body

    def _get_httpcore_pool_options(
        self: "Self", http2: bool = False
    ) -> "Dict[str, Any]":
        """Shared pool options for httpcore-based transports (Http2 and Async)."""
        options: "Dict[str, Any]" = {
            "http2": http2,
            "retries": 3,
        }

        socket_options: "Optional[List[Tuple[int, int, int | bytes]]]" = None

        if self.options["socket_options"] is not None:
            socket_options = self.options["socket_options"]

        if socket_options is None:
            socket_options = []

        used_options = {(o[0], o[1]) for o in socket_options}
        for default_option in KEEP_ALIVE_SOCKET_OPTIONS:
            if (default_option[0], default_option[1]) not in used_options:
                socket_options.append(default_option)

        if socket_options is not None:
            options["socket_options"] = socket_options

        ssl_context = ssl.create_default_context()
        ssl_context.load_verify_locations(
            self.options["ca_certs"]
            or os.environ.get("SSL_CERT_FILE")
            or os.environ.get("REQUESTS_CA_BUNDLE")
            or certifi.where()
        )
        cert_file = self.options["cert_file"] or os.environ.get("CLIENT_CERT_FILE")
        key_file = self.options["key_file"] or os.environ.get("CLIENT_KEY_FILE")
        if cert_file is not None:
            ssl_context.load_cert_chain(cert_file, key_file)

        options["ssl_context"] = ssl_context
        return options

    def _resolve_proxy(self: "Self") -> "Optional[str]":
        """Resolve proxy URL from options and environment. Returns proxy URL or None."""
        if self.parsed_dsn is None:
            return None

        no_proxy = self._in_no_proxy(self.parsed_dsn)
        proxy = None

        # try HTTPS first
        https_proxy = self.options["https_proxy"]
        if self.parsed_dsn.scheme == "https" and (https_proxy != ""):
            proxy = https_proxy or (not no_proxy and getproxies().get("https"))

        # maybe fallback to HTTP proxy
        http_proxy = self.options["http_proxy"]
        if not proxy and (http_proxy != ""):
            proxy = http_proxy or (not no_proxy and getproxies().get("http"))

        return proxy or None

    @property
    def _timeout_extensions(self: "Self") -> "Dict[str, Any]":
        return {
            "timeout": {
                "pool": self.TIMEOUT,
                "connect": self.TIMEOUT,
                "write": self.TIMEOUT,
                "read": self.TIMEOUT,
            }
        }

    def _get_pool_options(self: "Self") -> "Dict[str, Any]":
        raise NotImplementedError()

    def _in_no_proxy(self: "Self", parsed_dsn: "Dsn") -> bool:
        no_proxy = getproxies().get("no")
        if not no_proxy:
            return False
        for host in no_proxy.split(","):
            host = host.strip()
            if parsed_dsn.host.endswith(host) or parsed_dsn.netloc.endswith(host):
                return True
        return False

    def _make_pool(
        self: "Self",
    ) -> "Union[PoolManager, ProxyManager, httpcore.SOCKSProxy, httpcore.HTTPProxy, httpcore.ConnectionPool, httpcore.AsyncSOCKSProxy, httpcore.AsyncHTTPProxy, httpcore.AsyncConnectionPool]":
        raise NotImplementedError()

    def _request(
        self: "Self",
        method: str,
        endpoint_type: "EndpointType",
        body: "Any",
        headers: "Mapping[str, str]",
    ) -> "Union[urllib3.BaseHTTPResponse, httpcore.Response]":
        raise NotImplementedError()

    def kill(self: "Self") -> None:
        logger.debug("Killing HTTP transport")
        self._worker.kill()


# Keep BaseHttpTransport as an alias for backwards compatibility
# and for the sync transport implementation
class BaseHttpTransport(HttpTransportCore):
    """The base HTTP transport (synchronous)."""

    def _send_envelope(self: "Self", envelope: "Envelope") -> None:
        _prepared_envelope = self._prepare_envelope(envelope)
        if _prepared_envelope is not None:
            envelope, body, headers = _prepared_envelope
            self._send_request(
                body.getvalue(),
                headers=headers,
                endpoint_type=EndpointType.ENVELOPE,
                envelope=envelope,
            )
        return None

    def _send_request(
        self: "Self",
        body: bytes,
        headers: "Dict[str, str]",
        endpoint_type: "EndpointType",
        envelope: "Optional[Envelope]" = None,
    ) -> None:
        self._update_headers(headers)
        try:
            response = self._request(
                "POST",
                endpoint_type,
                body,
                headers,
            )
        except Exception:
            self._handle_request_error(envelope=envelope, loss_reason="network")
            raise
        try:
            self._handle_response(response=response, envelope=envelope)
        finally:
            response.close()

    def _create_worker(self: "Self", options: "Dict[str, Any]") -> "Worker":
        return BackgroundWorker(queue_size=options["transport_queue_size"])

    def _flush_client_reports(self: "Self", force: bool = False) -> None:
        client_report = self._fetch_pending_client_report(force=force, interval=60)
        if client_report is not None:
            self.capture_envelope(Envelope(items=[client_report]))

    def capture_envelope(
        self,
        envelope: "Envelope",
    ) -> None:
        def send_envelope_wrapper() -> None:
            with capture_internal_exceptions():
                self._send_envelope(envelope)
                self._flush_client_reports()

        if not self._worker.submit(send_envelope_wrapper):
            self.on_dropped_event("full_queue")
            for item in envelope.items:
                self.record_lost_event("queue_overflow", item=item)

    def flush(
        self: "Self",
        timeout: float,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> None:
        logger.debug("Flushing HTTP transport")

        if timeout > 0:
            self._worker.submit(lambda: self._flush_client_reports(force=True))
            self._worker.flush(timeout, callback)

    @staticmethod
    def _warn_hub_cls() -> None:
        """Convenience method to warn users about the deprecation of the `hub_cls` attribute."""
        warnings.warn(
            "The `hub_cls` attribute is deprecated and will be removed in a future release.",
            DeprecationWarning,
            stacklevel=3,
        )

    @property
    def hub_cls(self: "Self") -> "type[sentry_sdk.Hub]":
        """DEPRECATED: This attribute is deprecated and will be removed in a future release."""
        HttpTransport._warn_hub_cls()
        return self._hub_cls

    @hub_cls.setter
    def hub_cls(self: "Self", value: "type[sentry_sdk.Hub]") -> None:
        """DEPRECATED: This attribute is deprecated and will be removed in a future release."""
        HttpTransport._warn_hub_cls()
        self._hub_cls = value


class HttpTransport(BaseHttpTransport):
    if TYPE_CHECKING:
        _pool: "Union[PoolManager, ProxyManager]"

    def _get_pool_options(self: "Self") -> "Dict[str, Any]":
        num_pools = self.options.get("_experiments", {}).get("transport_num_pools")
        options = {
            "num_pools": 2 if num_pools is None else int(num_pools),
            "cert_reqs": "CERT_REQUIRED",
            "timeout": urllib3.Timeout(total=self.TIMEOUT),
        }

        socket_options: "Optional[List[Tuple[int, int, int | bytes]]]" = None

        if self.options["socket_options"] is not None:
            socket_options = self.options["socket_options"]

        if self.options["keep_alive"]:
            if socket_options is None:
                socket_options = []

            used_options = {(o[0], o[1]) for o in socket_options}
            for default_option in KEEP_ALIVE_SOCKET_OPTIONS:
                if (default_option[0], default_option[1]) not in used_options:
                    socket_options.append(default_option)

        if socket_options is not None:
            options["socket_options"] = socket_options

        options["ca_certs"] = (
            self.options["ca_certs"]  # User-provided bundle from the SDK init
            or os.environ.get("SSL_CERT_FILE")
            or os.environ.get("REQUESTS_CA_BUNDLE")
            or certifi.where()
        )

        options["cert_file"] = self.options["cert_file"] or os.environ.get(
            "CLIENT_CERT_FILE"
        )
        options["key_file"] = self.options["key_file"] or os.environ.get(
            "CLIENT_KEY_FILE"
        )

        return options

    def _make_pool(self: "Self") -> "Union[PoolManager, ProxyManager]":
        if self.parsed_dsn is None:
            raise ValueError("Cannot create HTTP-based transport without valid DSN")

        proxy = None
        no_proxy = self._in_no_proxy(self.parsed_dsn)

        # try HTTPS first
        https_proxy = self.options["https_proxy"]
        if self.parsed_dsn.scheme == "https" and (https_proxy != ""):
            proxy = https_proxy or (not no_proxy and getproxies().get("https"))

        # maybe fallback to HTTP proxy
        http_proxy = self.options["http_proxy"]
        if not proxy and (http_proxy != ""):
            proxy = http_proxy or (not no_proxy and getproxies().get("http"))

        opts = self._get_pool_options()

        if proxy:
            proxy_headers = self.options["proxy_headers"]
            if proxy_headers:
                opts["proxy_headers"] = proxy_headers

            if proxy.startswith("socks"):
                use_socks_proxy = True
                try:
                    # Check if PySocks dependency is available
                    from urllib3.contrib.socks import SOCKSProxyManager
                except ImportError:
                    use_socks_proxy = False
                    logger.warning(
                        "You have configured a SOCKS proxy (%s) but support for SOCKS proxies is not installed. Disabling proxy support. Please add `PySocks` (or `urllib3` with the `[socks]` extra) to your dependencies.",
                        proxy,
                    )

                if use_socks_proxy:
                    return SOCKSProxyManager(proxy, **opts)
                else:
                    return urllib3.PoolManager(**opts)
            else:
                return urllib3.ProxyManager(proxy, **opts)
        else:
            return urllib3.PoolManager(**opts)

    def _request(
        self: "Self",
        method: str,
        endpoint_type: "EndpointType",
        body: "Any",
        headers: "Mapping[str, str]",
    ) -> "urllib3.BaseHTTPResponse":
        return self._pool.request(
            method,
            self._auth.get_api_url(endpoint_type),
            body=body,
            headers=headers,
        )


class AsyncHttpTransport(HttpTransportCore):
    def __init__(self: "Self", options: "Dict[str, Any]") -> None:
        if not ASYNC_TRANSPORT_AVAILABLE:
            raise RuntimeError(
                "AsyncHttpTransport requires httpcore[asyncio]. "
                "Install it with: pip install sentry-sdk[asyncio]"
            )
        super().__init__(options)
        # Requires event loop at init time
        self.loop = asyncio.get_running_loop()

    def _create_worker(self: "Self", options: "Dict[str, Any]") -> "Worker":
        return AsyncWorker(queue_size=options["transport_queue_size"])

    def _get_header_value(
        self: "Self", response: "Any", header: str
    ) -> "Optional[str]":
        return _get_httpcore_header_value(response, header)

    async def _send_envelope(self: "Self", envelope: "Envelope") -> None:
        _prepared_envelope = self._prepare_envelope(envelope)
        if _prepared_envelope is not None:
            envelope, body, headers = _prepared_envelope
            await self._send_request(
                body.getvalue(),
                headers=headers,
                endpoint_type=EndpointType.ENVELOPE,
                envelope=envelope,
            )
        return None

    async def _send_request(
        self: "Self",
        body: bytes,
        headers: "Dict[str, str]",
        endpoint_type: "EndpointType",
        envelope: "Optional[Envelope]" = None,
    ) -> None:
        self._update_headers(headers)
        try:
            response = await self._request(
                "POST",
                endpoint_type,
                body,
                headers,
            )
        except Exception:
            self._handle_request_error(envelope=envelope, loss_reason="network")
            raise
        try:
            self._handle_response(response=response, envelope=envelope)
        finally:
            await response.aclose()

    async def _request(  # type: ignore[override]
        self: "Self",
        method: str,
        endpoint_type: "EndpointType",
        body: "Any",
        headers: "Mapping[str, str]",
    ) -> "httpcore.Response":
        return await self._pool.request(  # type: ignore[misc,unused-ignore]
            method,
            self._auth.get_api_url(endpoint_type),
            content=body,
            headers=headers,  # type: ignore[arg-type,unused-ignore]
            extensions=self._timeout_extensions,
        )

    async def _flush_client_reports(self: "Self", force: bool = False) -> None:
        client_report = self._fetch_pending_client_report(force=force, interval=60)
        if client_report is not None:
            self.capture_envelope(Envelope(items=[client_report]))

    def _capture_envelope(self: "Self", envelope: "Envelope") -> None:
        async def send_envelope_wrapper() -> None:
            with capture_internal_exceptions():
                await self._send_envelope(envelope)
                await self._flush_client_reports()

        if not self._worker.submit(send_envelope_wrapper):
            self.on_dropped_event("full_queue")
            for item in envelope.items:
                self.record_lost_event("queue_overflow", item=item)

    def capture_envelope(self: "Self", envelope: "Envelope") -> None:
        # Synchronous entry point
        if self.loop and self.loop.is_running():
            self.loop.call_soon_threadsafe(self._capture_envelope, envelope)
        else:
            # The event loop is no longer running
            logger.warning("Async Transport is not running in an event loop.")
            self.on_dropped_event("internal_sdk_error")
            for item in envelope.items:
                self.record_lost_event("internal_sdk_error", item=item)

    def flush(  # type: ignore[override]
        self: "Self",
        timeout: float,
        callback: "Optional[Callable[[int, float], None]]" = None,
    ) -> "Optional[asyncio.Task[None]]":
        logger.debug("Flushing HTTP transport")

        if timeout > 0:
            self._worker.submit(lambda: self._flush_client_reports(force=True))
            return self._worker.flush(timeout, callback)  # type: ignore[func-returns-value]
        return None

    def _get_pool_options(self: "Self") -> "Dict[str, Any]":
        return self._get_httpcore_pool_options(
            http2=HTTP2_ENABLED
            and self.parsed_dsn is not None
            and self.parsed_dsn.scheme == "https"
        )

    def _make_pool(
        self: "Self",
    ) -> "Union[httpcore.AsyncSOCKSProxy, httpcore.AsyncHTTPProxy, httpcore.AsyncConnectionPool]":
        if self.parsed_dsn is None:
            raise ValueError("Cannot create HTTP-based transport without valid DSN")

        proxy = self._resolve_proxy()
        opts = self._get_pool_options()

        if proxy:
            proxy_headers = self.options["proxy_headers"]
            if proxy_headers:
                opts["proxy_headers"] = proxy_headers

            if proxy.startswith("socks"):
                try:
                    socks_opts = opts.copy()
                    if "socket_options" in socks_opts:
                        socket_options = socks_opts.pop("socket_options")
                        if socket_options:
                            logger.warning(
                                "You have defined socket_options but using a SOCKS proxy which doesn't support these. We'll ignore socket_options."
                            )
                    return httpcore.AsyncSOCKSProxy(proxy_url=proxy, **socks_opts)
                except RuntimeError:
                    logger.warning(
                        "You have configured a SOCKS proxy (%s) but support for SOCKS proxies is not installed. Disabling proxy support.",
                        proxy,
                    )
            else:
                return httpcore.AsyncHTTPProxy(proxy_url=proxy, **opts)

        return httpcore.AsyncConnectionPool(**opts)

    def kill(self: "Self") -> "Optional[asyncio.Task[None]]":  # type: ignore[override]
        logger.debug("Killing HTTP transport")
        self._worker.kill()
        try:
            # Return the pool cleanup task so caller can await it if needed
            with mark_sentry_task_internal():
                return self.loop.create_task(self._pool.aclose())  # type: ignore[union-attr,unused-ignore]
        except RuntimeError:
            logger.warning("Event loop not running, aborting kill.")
            return None


if not HTTP2_ENABLED:
    # Sorry, no Http2Transport for you
    class Http2Transport(HttpTransport):
        def __init__(self: "Self", options: "Dict[str, Any]") -> None:
            super().__init__(options)
            logger.warning(
                "You tried to use HTTP2Transport but don't have httpcore[http2] installed. Falling back to HTTPTransport."
            )

else:

    class Http2Transport(BaseHttpTransport):  # type: ignore
        """The HTTP2 transport based on httpcore."""

        TIMEOUT = 15

        if TYPE_CHECKING:
            _pool: """Union[
                httpcore.SOCKSProxy, httpcore.HTTPProxy, httpcore.ConnectionPool
            ]"""

        def _get_header_value(
            self: "Self", response: "httpcore.Response", header: str
        ) -> "Optional[str]":
            return _get_httpcore_header_value(response, header)

        def _request(
            self: "Self",
            method: str,
            endpoint_type: "EndpointType",
            body: "Any",
            headers: "Mapping[str, str]",
        ) -> "httpcore.Response":
            response = self._pool.request(
                method,
                self._auth.get_api_url(endpoint_type),
                content=body,
                headers=headers,  # type: ignore[arg-type,unused-ignore]
                extensions=self._timeout_extensions,
            )
            return response

        def _get_pool_options(self: "Self") -> "Dict[str, Any]":
            return self._get_httpcore_pool_options(
                http2=self.parsed_dsn is not None and self.parsed_dsn.scheme == "https"
            )

        def _make_pool(
            self: "Self",
        ) -> "Union[httpcore.SOCKSProxy, httpcore.HTTPProxy, httpcore.ConnectionPool]":
            if self.parsed_dsn is None:
                raise ValueError("Cannot create HTTP-based transport without valid DSN")

            proxy = self._resolve_proxy()
            opts = self._get_pool_options()

            if proxy:
                proxy_headers = self.options["proxy_headers"]
                if proxy_headers:
                    opts["proxy_headers"] = proxy_headers

                if proxy.startswith("socks"):
                    try:
                        if "socket_options" in opts:
                            socket_options = opts.pop("socket_options")
                            if socket_options:
                                logger.warning(
                                    "You have defined socket_options but using a SOCKS proxy which doesn't support these. We'll ignore socket_options."
                                )
                        return httpcore.SOCKSProxy(proxy_url=proxy, **opts)
                    except RuntimeError:
                        logger.warning(
                            "You have configured a SOCKS proxy (%s) but support for SOCKS proxies is not installed. Disabling proxy support.",
                            proxy,
                        )
                else:
                    return httpcore.HTTPProxy(proxy_url=proxy, **opts)

            return httpcore.ConnectionPool(**opts)


class _EnvelopePrinterTransport(Transport):
    """Wraps another transport, printing envelope contents to the SDK debug logger before sending."""

    def __init__(self, transport: "Transport") -> None:
        Transport.__init__(self, options=transport.options)
        self._inner = transport
        self.parsed_dsn = transport.parsed_dsn

        self.envelope_logger = logging.getLogger("sentry_sdk.envelopes")
        self.envelope_logger.setLevel(logging.INFO)
        self.envelope_logger.propagate = False
        if not self.envelope_logger.handlers:
            handler = logging.StreamHandler()
            handler.setLevel(logging.INFO)
            handler.setFormatter(logging.Formatter("%(message)s"))
            self.envelope_logger.addHandler(handler)

    @property  # type: ignore[misc]
    def __class__(self) -> type:
        return self._inner.__class__

    def capture_envelope(self, envelope: "Envelope") -> None:
        try:
            self.envelope_logger.info("--- Sentry Envelope ---")
            self.envelope_logger.info(
                "Headers: %s", json.dumps(envelope.headers, indent=2, default=str)
            )
            for item in envelope.items:
                self.envelope_logger.info("  Item type: %s", item.type)
                self.envelope_logger.info(
                    "  Item headers: %s",
                    json.dumps(item.headers, indent=2, default=str),
                )
                try:
                    payload = json.loads(item.get_bytes())
                    self.envelope_logger.info(
                        "  Payload:\n%s",
                        json.dumps(payload, indent=2, default=str),
                    )
                except (ValueError, TypeError):
                    self.envelope_logger.info(
                        "  Payload: <binary %d bytes>",
                        len(item.get_bytes()),
                    )
            self.envelope_logger.info("--- End Envelope ---")
        except Exception:
            pass

        self._inner.capture_envelope(envelope)

    def flush(
        self,
        timeout: float,
        callback: "Optional[Any]" = None,
    ) -> "Any":
        return self._inner.flush(timeout, callback)

    def kill(self) -> "Any":
        return self._inner.kill()

    def record_lost_event(
        self,
        reason: str,
        data_category: "Optional[EventDataCategory]" = None,
        item: "Optional[Item]" = None,
        *,
        quantity: int = 1,
    ) -> None:
        self._inner.record_lost_event(reason, data_category, item, quantity=quantity)

    def is_healthy(self) -> bool:
        return self._inner.is_healthy()

    def __getattr__(self, name: str) -> "Any":
        return getattr(self._inner, name)


class _FunctionTransport(Transport):
    """
    DEPRECATED: Users wishing to provide a custom transport should subclass
    the Transport class, rather than providing a function.
    """

    def __init__(
        self,
        func: "Callable[[Event], None]",
    ) -> None:
        Transport.__init__(self)
        self._func = func

    def capture_event(
        self,
        event: "Event",
    ) -> None:
        self._func(event)
        return None

    def capture_envelope(self, envelope: "Envelope") -> None:
        # Since function transports expect to be called with an event, we need
        # to iterate over the envelope and call the function for each event, via
        # the deprecated capture_event method.
        event = envelope.get_event()
        if event is not None:
            self.capture_event(event)


def make_transport(options: "Dict[str, Any]") -> "Optional[Transport]":
    ref_transport = options["transport"]

    use_http2_transport = options.get("_experiments", {}).get("transport_http2", False)
    use_async_transport = options.get("_experiments", {}).get("transport_async", False)
    async_integration = any(
        integration.__class__.__name__ == "AsyncioIntegration"
        for integration in options.get("integrations") or []
    )

    # By default, we use the http transport class
    transport_cls: "Type[Transport]" = (
        Http2Transport if use_http2_transport else HttpTransport
    )

    if use_async_transport and ASYNC_TRANSPORT_AVAILABLE:
        try:
            asyncio.get_running_loop()
            if async_integration:
                if use_http2_transport:
                    logger.warning(
                        "HTTP/2 transport is not supported with async transport. "
                        "Ignoring transport_http2 experiment."
                    )
                transport_cls = AsyncHttpTransport
            else:
                logger.warning(
                    "You tried to use AsyncHttpTransport but the AsyncioIntegration is not enabled. Falling back to sync transport."
                )
        except RuntimeError:
            # No event loop running, fall back to sync transport
            logger.warning("No event loop running, falling back to sync transport.")
    elif use_async_transport:
        logger.warning(
            "You tried to use AsyncHttpTransport but don't have httpcore[asyncio] installed. Falling back to sync transport."
        )

    transport: "Optional[Transport]" = None

    if isinstance(ref_transport, Transport):
        transport = ref_transport
    elif isinstance(ref_transport, type) and issubclass(ref_transport, Transport):
        transport_cls = ref_transport
    elif callable(ref_transport):
        warnings.warn(
            "Function transports are deprecated and will be removed in a future release."
            "Please provide a Transport instance or subclass, instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        transport = _FunctionTransport(ref_transport)

    # if a transport class is given only instantiate it if the dsn is not
    # empty or None
    if transport is None and options["dsn"]:
        transport = transport_cls(options)

    if transport is not None and os.environ.get(
        "SENTRY_PRINT_ENVELOPES", ""
    ).lower() in ("1", "true", "yes"):
        transport = _EnvelopePrinterTransport(transport)

    return transport



import functools
import inspect
import sys
import threading
import weakref
from importlib import import_module

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA, SPANNAME
from sentry_sdk.integrations import (
    _DEFAULT_FAILED_REQUEST_STATUS_CODES,
    DidNotEnable,
    Integration,
    _check_minimum_version,
)
from sentry_sdk.integrations._wsgi_common import (
    DEFAULT_HTTP_METHODS_TO_CAPTURE,
    RequestExtractor,
)
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware
from sentry_sdk.scope import add_global_event_processor, should_send_default_pii
from sentry_sdk.serializer import add_global_repr_processor, add_repr_sequence_type
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import SOURCE_FOR_STYLE, TransactionSource
from sentry_sdk.tracing_utils import (
    add_query_source,
    has_span_streaming_enabled,
    record_sql_queries,
)
from sentry_sdk.utils import (
    CONTEXTVARS_ERROR_MESSAGE,
    HAS_REAL_CONTEXTVARS,
    SENSITIVE_DATA_SUBSTITUTE,
    AnnotatedValue,
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    has_data_collection_enabled,
    logger,
    transaction_from_function,
    walk_exception_chain,
)

try:
    from django import VERSION as DJANGO_VERSION
    from django.conf import settings
    from django.conf import settings as django_settings
    from django.core import signals
    from django.utils.functional import SimpleLazyObject

    try:
        from django.urls import resolve
    except ImportError:
        from django.core.urlresolvers import resolve

    try:
        from django.urls import Resolver404
    except ImportError:
        from django.core.urlresolvers import Resolver404

    # Only available in Django 3.0+
    try:
        from django.core.handlers.asgi import ASGIRequest
    except Exception:
        ASGIRequest = None

except ImportError:
    raise DidNotEnable("Django not installed")

from sentry_sdk.integrations.django.middleware import patch_django_middlewares
from sentry_sdk.integrations.django.signals_handlers import patch_signals
from sentry_sdk.integrations.django.tasks import patch_tasks
from sentry_sdk.integrations.django.templates import (
    get_template_frame_from_exception,
    patch_templates,
)
from sentry_sdk.integrations.django.transactions import LEGACY_RESOLVER
from sentry_sdk.integrations.django.views import patch_views

if DJANGO_VERSION[:2] > (1, 8):
    from sentry_sdk.integrations.django.caching import patch_caching
else:
    patch_caching = None  # type: ignore

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Set
    from typing import Any, Callable, Dict, List, Optional, Union

    from django.core.handlers.wsgi import WSGIRequest
    from django.http.request import QueryDict
    from django.http.response import HttpResponse
    from django.utils.datastructures import MultiValueDict

    from sentry_sdk._types import (
        Event,
        EventProcessor,
        ExcInfo,
        Hint,
        NotImplementedType,
    )
    from sentry_sdk.integrations.wsgi import _ScopedResponse
    from sentry_sdk.traces import StreamedSpan
    from sentry_sdk.tracing import Span


if DJANGO_VERSION < (1, 10):

    def is_authenticated(request_user: "Any") -> bool:
        return request_user.is_authenticated()

else:

    def is_authenticated(request_user: "Any") -> bool:
        return request_user.is_authenticated


TRANSACTION_STYLE_VALUES = ("function_name", "url")


class DjangoIntegration(Integration):
    """
    Auto instrument a Django application.

    :param transaction_style: How to derive transaction names. Either `"function_name"` or `"url"`. Defaults to `"url"`.
    :param middleware_spans: Whether to create spans for middleware. Defaults to `False`.
    :param signals_spans: Whether to create spans for signals. Defaults to `True`.
    :param signals_denylist: A list of signals to ignore when creating spans.
    :param cache_spans: Whether to create spans for cache operations. Defaults to `False`.
    :param failed_request_status_codes: Which HTTP error responses to report to Sentry.
        Django answers some exceptions itself instead of failing: `raise Http404` gets
        the user a 404 page, `PermissionDenied` a 403. Those are reported only if their
        status code is in this set, which defaults to the 5xx range. Exceptions Django
        gives up on end in a 500 and are always reported.
    """

    identifier = "django"
    origin = f"auto.http.{identifier}"
    origin_db = f"auto.db.{identifier}"

    transaction_style = ""
    middleware_spans: "Optional[bool]" = None
    signals_spans: "Optional[bool]" = None
    cache_spans: "Optional[bool]" = None
    signals_denylist: "list[signals.Signal]" = []

    def __init__(
        self,
        transaction_style: str = "url",
        middleware_spans: bool = False,
        signals_spans: bool = True,
        cache_spans: bool = False,
        db_transaction_spans: bool = False,
        signals_denylist: "Optional[list[signals.Signal]]" = None,
        http_methods_to_capture: "tuple[str, ...]" = DEFAULT_HTTP_METHODS_TO_CAPTURE,
        *,
        failed_request_status_codes: "Set[int]" = _DEFAULT_FAILED_REQUEST_STATUS_CODES,
    ) -> None:
        if transaction_style not in TRANSACTION_STYLE_VALUES:
            raise ValueError(
                "Invalid value for transaction_style: %s (must be in %s)"
                % (transaction_style, TRANSACTION_STYLE_VALUES)
            )
        self.transaction_style = transaction_style
        self.middleware_spans = middleware_spans

        self.signals_spans = signals_spans
        self.signals_denylist = signals_denylist or []

        self.cache_spans = cache_spans
        self.db_transaction_spans = db_transaction_spans

        self.http_methods_to_capture = tuple(map(str.upper, http_methods_to_capture))

        self.failed_request_status_codes = failed_request_status_codes

    @staticmethod
    def setup_once() -> None:
        _check_minimum_version(DjangoIntegration, DJANGO_VERSION)

        install_sql_hook()
        # Patch in our custom middleware.

        # logs an error for every 500
        ignore_logger("django.server")
        ignore_logger("django.request")

        from django.core.handlers.wsgi import WSGIHandler

        old_app = WSGIHandler.__call__

        @ensure_integration_enabled(DjangoIntegration, old_app)
        def sentry_patched_wsgi_handler(
            self: "Any", environ: "Dict[str, str]", start_response: "Callable[..., Any]"
        ) -> "_ScopedResponse":
            bound_old_app = old_app.__get__(self, WSGIHandler)

            from django.conf import settings

            use_x_forwarded_for = settings.USE_X_FORWARDED_HOST

            integration = sentry_sdk.get_client().get_integration(DjangoIntegration)

            middleware = SentryWsgiMiddleware(
                bound_old_app,
                use_x_forwarded_for,
                span_origin=DjangoIntegration.origin,
                http_methods_to_capture=(
                    integration.http_methods_to_capture
                    if integration
                    else DEFAULT_HTTP_METHODS_TO_CAPTURE
                ),
            )
            return middleware(environ, start_response)

        WSGIHandler.__call__ = sentry_patched_wsgi_handler

        _patch_get_response()

        _patch_django_asgi_handler()

        _patch_response_for_exception()

        signals.got_request_exception.connect(_got_request_exception)

        @add_global_event_processor
        def process_django_templates(
            event: "Event", hint: "Optional[Hint]"
        ) -> "Optional[Event]":
            if hint is None:
                return event

            exc_info = hint.get("exc_info", None)

            if exc_info is None:
                return event

            exception = event.get("exception", None)

            if exception is None:
                return event

            values = exception.get("values", None)

            if values is None:
                return event

            for exception, (_, exc_value, _) in zip(
                reversed(values), walk_exception_chain(exc_info)
            ):
                frame = get_template_frame_from_exception(exc_value)
                if frame is not None:
                    frames = exception.get("stacktrace", {}).get("frames", [])

                    for i in reversed(range(len(frames))):
                        f = frames[i]
                        if (
                            f.get("function") in ("Parser.parse", "parse", "render")
                            and f.get("module") == "django.template.base"
                        ):
                            i += 1
                            break
                    else:
                        i = len(frames)

                    frames.insert(i, frame)

            return event

        @add_global_repr_processor
        def _django_queryset_repr(
            value: "Any", hint: "Dict[str, Any]"
        ) -> "Union[NotImplementedType, str]":
            try:
                # Django 1.6 can fail to import `QuerySet` when Django settings
                # have not yet been initialized.
                #
                # If we fail to import, return `NotImplemented`. It's at least
                # unlikely that we have a query set in `value` when importing
                # `QuerySet` fails.
                from django.db.models.query import QuerySet
            except Exception:
                return NotImplemented

            if not isinstance(value, QuerySet) or value._result_cache:
                return NotImplemented

            return "<%s from %s at 0x%x>" % (
                value.__class__.__name__,
                value.__module__,
                id(value),
            )

        _patch_channels()
        patch_django_middlewares()
        patch_views()
        patch_templates()
        patch_signals()
        patch_tasks()
        add_template_context_repr_sequence()

        if patch_caching is not None:
            patch_caching()


_DRF_PATCHED = False
_DRF_PATCH_LOCK = threading.Lock()


def _patch_drf() -> None:
    """
    Patch Django Rest Framework for more/better request data. DRF's request
    type is a wrapper around Django's request type. The attribute we're
    interested in is `request.data`, which is a cached property containing a
    parsed request body. Reading a request body from that property is more
    reliable than reading from any of Django's own properties, as those don't
    hold payloads in memory and therefore can only be accessed once.

    We patch the Django request object to include a weak backreference to the
    DRF request object, such that we can later use either in
    `DjangoRequestExtractor`.

    We also patch DRF's authentication to create a span, so that the work done
    by the configured authentication classes (which often involves database
    queries) doesn't show up as part of the view itself.

    This function is not called directly on SDK setup, because importing almost
    any part of Django Rest Framework will try to access Django settings (where
    `sentry_sdk.init()` might be called from in the first place). Instead we
    run this function on every request and do the patching on the first
    request.
    """

    global _DRF_PATCHED

    if _DRF_PATCHED:
        # Double-checked locking
        return

    with _DRF_PATCH_LOCK:
        if _DRF_PATCHED:
            return

        # We set this regardless of whether the code below succeeds or fails.
        # There is no point in trying to patch again on the next request.
        _DRF_PATCHED = True

        with capture_internal_exceptions():
            try:
                from rest_framework.views import APIView  # type: ignore
            except ImportError:
                pass
            else:
                old_drf_initial = APIView.initial

                def sentry_patched_drf_initial(
                    self: "APIView", request: "Any", *args: "Any", **kwargs: "Any"
                ) -> "Any":
                    with capture_internal_exceptions():
                        request._request._sentry_drf_request_backref = weakref.ref(
                            request
                        )
                        pass
                    return old_drf_initial(self, request, *args, **kwargs)

                APIView.initial = sentry_patched_drf_initial

        with capture_internal_exceptions():
            try:
                from rest_framework.request import Request  # type: ignore
            except ImportError:
                pass
            else:
                old_drf_authenticate = Request._authenticate

                def sentry_patched_drf_authenticate(self: "Request") -> "Any":
                    client = sentry_sdk.get_client()
                    integration = client.get_integration(DjangoIntegration)
                    # Nothing to time if there are no authenticators configured
                    # for this view.
                    if integration is None or not getattr(self, "authenticators", None):
                        return old_drf_authenticate(self)

                    if has_span_streaming_enabled(client.options):
                        if sentry_sdk.traces.get_current_span() is None:
                            return old_drf_authenticate(self)
                        with sentry_sdk.traces.start_span(
                            name="authenticate",
                            attributes={
                                "sentry.op": OP.VIEW_AUTHENTICATE,
                                "sentry.origin": DjangoIntegration.origin,
                            },
                        ):
                            return old_drf_authenticate(self)
                    else:
                        with sentry_sdk.start_span(
                            op=OP.VIEW_AUTHENTICATE,
                            name="authenticate",
                            origin=DjangoIntegration.origin,
                        ):
                            return old_drf_authenticate(self)

                Request._authenticate = sentry_patched_drf_authenticate


def _patch_channels() -> None:
    try:
        from channels.http import AsgiHandler  # type: ignore
    except ImportError:
        return

    if not HAS_REAL_CONTEXTVARS:
        # We better have contextvars or we're going to leak state between
        # requests.
        #
        # We cannot hard-raise here because channels may not be used at all in
        # the current process. That is the case when running traditional WSGI
        # workers in gunicorn+gevent and the websocket stuff in a separate
        # process.
        logger.warning(
            "We detected that you are using Django channels 2.0."
            + CONTEXTVARS_ERROR_MESSAGE
        )

    from sentry_sdk.integrations.django.asgi import patch_channels_asgi_handler_impl

    patch_channels_asgi_handler_impl(AsgiHandler)


def _patch_django_asgi_handler() -> None:
    try:
        from django.core.handlers.asgi import ASGIHandler
    except ImportError:
        return

    if not HAS_REAL_CONTEXTVARS:
        # We better have contextvars or we're going to leak state between
        # requests.
        #
        # We cannot hard-raise here because Django's ASGI stuff may not be used
        # at all.
        logger.warning(
            "We detected that you are using Django 3." + CONTEXTVARS_ERROR_MESSAGE
        )

    from sentry_sdk.integrations.django.asgi import patch_django_asgi_handler_impl

    patch_django_asgi_handler_impl(ASGIHandler)


def _set_transaction_name_and_source(
    scope: "sentry_sdk.Scope", transaction_style: str, request: "WSGIRequest"
) -> None:
    try:
        transaction_name = None
        if transaction_style == "function_name":
            fn = resolve(request.path).func
            transaction_name = transaction_from_function(getattr(fn, "view_class", fn))

        elif transaction_style == "url":
            if hasattr(request, "urlconf"):
                transaction_name = LEGACY_RESOLVER.resolve(
                    request.path_info, urlconf=request.urlconf
                )
            else:
                transaction_name = LEGACY_RESOLVER.resolve(request.path_info)

        if transaction_name is None:
            transaction_name = request.path_info
            source = TransactionSource.URL
        else:
            source = SOURCE_FOR_STYLE[transaction_style]

        scope.set_transaction_name(
            transaction_name,
            source=source,
        )
    except Resolver404:
        urlconf = import_module(settings.ROOT_URLCONF)
        # This exception only gets thrown when transaction_style is `function_name`
        # So we don't check here what style is configured
        if hasattr(urlconf, "handler404"):
            handler = urlconf.handler404
            if isinstance(handler, str):
                scope.transaction = handler
            else:
                scope.transaction = transaction_from_function(
                    getattr(handler, "view_class", handler)
                )
    except Exception:
        pass


def _before_get_response(request: "WSGIRequest") -> None:
    integration = sentry_sdk.get_client().get_integration(DjangoIntegration)
    if integration is None:
        return

    _patch_drf()

    scope = sentry_sdk.get_current_scope()
    # Rely on WSGI middleware to start a trace
    _set_transaction_name_and_source(scope, integration.transaction_style, request)

    scope.add_event_processor(
        _make_wsgi_request_event_processor(weakref.ref(request), integration)
    )


def _attempt_resolve_again(
    request: "WSGIRequest", scope: "sentry_sdk.Scope", transaction_style: str
) -> None:
    """
    Some django middlewares overwrite request.urlconf
    so we need to respect that contract,
    so we try to resolve the url again.
    """
    if not hasattr(request, "urlconf"):
        return

    _set_transaction_name_and_source(scope, transaction_style, request)


def _get_user_from_request_and_set_on_scope(request: "WSGIRequest") -> None:
    user = getattr(request, "user", None)

    # Evaluating a SimpleLazyObject in an async view can raise django.core.exceptions.SynchronousOnlyOperation.
    # Exit early if the user has not been materialized yet.
    is_lazy = isinstance(user, SimpleLazyObject)
    if is_lazy and hasattr(request, "_cached_user"):
        user = request._cached_user
    elif is_lazy:
        return

    if user is None or not is_authenticated(user):
        return

    user_info = {}
    try:
        user_info["id"] = str(user.pk)
    except Exception:
        pass

    try:
        user_info["email"] = user.email
    except Exception:
        pass

    try:
        user_info["username"] = user.get_username()
    except Exception:
        pass

    sentry_sdk.set_user(user_info)


def _after_get_response(request: "WSGIRequest") -> None:
    client = sentry_sdk.get_client()
    integration = client.get_integration(DjangoIntegration)
    if integration is None:
        return

    if integration.transaction_style == "url":
        scope = sentry_sdk.get_current_scope()
        _attempt_resolve_again(request, scope, integration.transaction_style)

    span_streaming = has_span_streaming_enabled(client.options)
    if span_streaming:
        if has_data_collection_enabled(client.options):
            if client.options["data_collection"]["user_info"]:
                _get_user_from_request_and_set_on_scope(request)
        elif should_send_default_pii():
            _get_user_from_request_and_set_on_scope(request)


def _patch_get_response() -> None:
    """
    patch get_response, because at that point we have the Django request object
    """
    from django.core.handlers.base import BaseHandler

    old_get_response = BaseHandler.get_response

    def sentry_patched_get_response(
        self: "Any", request: "WSGIRequest"
    ) -> "Union[HttpResponse, BaseException]":
        _before_get_response(request)
        rv = old_get_response(self, request)
        _after_get_response(request)
        return rv

    BaseHandler.get_response = sentry_patched_get_response

    if hasattr(BaseHandler, "get_response_async"):
        from sentry_sdk.integrations.django.asgi import patch_get_response_async

        patch_get_response_async(BaseHandler, _before_get_response)


def _make_wsgi_request_event_processor(
    weak_request: "Callable[[], WSGIRequest]", integration: "DjangoIntegration"
) -> "EventProcessor":
    def wsgi_request_event_processor(event: "Event", hint: "dict[str, Any]") -> "Event":
        # if the request is gone we are fine not logging the data from
        # it.  This might happen if the processor is pushed away to
        # another thread.
        request = weak_request()
        if request is None:
            return event

        django_3 = ASGIRequest is not None
        if django_3 and type(request) == ASGIRequest:
            # We have a `asgi_request_event_processor` for this.
            return event

        with capture_internal_exceptions():
            DjangoRequestExtractor(request).extract_into_event(event)

        client_options = sentry_sdk.get_client().options
        if has_data_collection_enabled(client_options):
            if client_options["data_collection"]["user_info"]:
                with capture_internal_exceptions():
                    _set_user_info(request, event)
        elif should_send_default_pii():
            with capture_internal_exceptions():
                _set_user_info(request, event)

        return event

    return wsgi_request_event_processor


def _got_request_exception(request: "WSGIRequest" = None, **kwargs: "Any") -> None:
    client = sentry_sdk.get_client()
    integration = client.get_integration(DjangoIntegration)
    if integration is None:
        return

    # Record that this exception is reported, so `_patch_response_for_exception`
    # doesn't report it a second time.
    with capture_internal_exceptions():
        request._sentry_exception_reported = True

    _capture_exception(sys.exc_info(), request, integration, handled=False)


def _capture_exception(
    exc_info: "Union[BaseException, ExcInfo]",
    request: "Optional[WSGIRequest]",
    integration: "DjangoIntegration",
    handled: bool,
) -> None:
    if request is not None and integration.transaction_style == "url":
        scope = sentry_sdk.get_current_scope()
        _attempt_resolve_again(request, scope, integration.transaction_style)

    event, hint = event_from_exception(
        exc_info,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "django", "handled": handled},
    )
    sentry_sdk.capture_event(event, hint=hint)


def _patch_response_for_exception() -> None:
    """
    Report the errors Django answers itself.

    Django deals with every exception in one function, which boils down to:

        if isinstance(exc, Http404):             return <404 page>
        if isinstance(exc, PermissionDenied):    return <403 page>
        if isinstance(exc, SuspiciousOperation): return <400 page>
        got_request_exception.send(...)          # Django gives up
        return <500 page>

    We only ever listened to that signal, so we heard about the exceptions Django
    gives up on and about nothing else. Wrapping the function lets us see the rest
    too, along with the status code Django picked for them.
    """
    try:
        from django.core.handlers import exception as exception_handler
    except ImportError:
        # Django < 1.10 does this in `BaseHandler`, nothing to patch here
        return

    old_response_for_exception = getattr(
        exception_handler, "response_for_exception", None
    )
    if old_response_for_exception is None:
        return

    @functools.wraps(old_response_for_exception)
    def sentry_patched_response_for_exception(
        request: "WSGIRequest", exc: Exception
    ) -> "HttpResponse":
        integration = sentry_sdk.get_client().get_integration(DjangoIntegration)
        if integration is None:
            return old_response_for_exception(request, exc)

        # Clear the flag before delegating. The same request can reach this
        # function twice: first when the view raises, then again if a middleware
        # raises while handing the response back out. Without the reset, the
        # first exception would keep the second one from being reported.
        with capture_internal_exceptions():
            request._sentry_exception_reported = False

        response = old_response_for_exception(request, exc)

        # The flag is set when Django gives up on the exception and fires
        # `got_request_exception`, which means we reported it already.
        if not getattr(request, "_sentry_exception_reported", False):
            status_code = getattr(response, "status_code", None)
            if status_code in integration.failed_request_status_codes:
                _capture_exception(exc, request, integration, handled=True)

        return response

    exception_handler.response_for_exception = sentry_patched_response_for_exception


class DjangoRequestExtractor(RequestExtractor):
    def __init__(self, request: "Union[WSGIRequest, ASGIRequest]") -> None:
        try:
            drf_request = request._sentry_drf_request_backref()
            if drf_request is not None:
                request = drf_request
        except AttributeError:
            pass
        self.request = request

    def env(self) -> "Dict[str, str]":
        return self.request.META

    def cookies(self) -> "Dict[str, Union[str, AnnotatedValue]]":
        privacy_cookies = [
            django_settings.CSRF_COOKIE_NAME,
            django_settings.SESSION_COOKIE_NAME,
        ]

        clean_cookies: "Dict[str, Union[str, AnnotatedValue]]" = {}
        for key, val in self.request.COOKIES.items():
            if key in privacy_cookies:
                clean_cookies[key] = SENSITIVE_DATA_SUBSTITUTE
            else:
                clean_cookies[key] = val

        return clean_cookies

    def raw_data(self) -> bytes:
        return self.request.body

    def form(self) -> "QueryDict":
        return self.request.POST

    def files(self) -> "MultiValueDict":
        return self.request.FILES

    def size_of_file(self, file: "Any") -> int:
        return file.size

    def parsed_body(self) -> "Optional[Dict[str, Any]]":
        try:
            return self.request.data
        except Exception:
            return RequestExtractor.parsed_body(self)


def _set_user_info(request: "WSGIRequest", event: "Event") -> None:
    user_info = event.setdefault("user", {})

    user = getattr(request, "user", None)

    if user is None or not is_authenticated(user):
        return

    try:
        user_info.setdefault("id", str(user.pk))
    except Exception:
        pass

    try:
        user_info.setdefault("email", user.email)
    except Exception:
        pass

    try:
        user_info.setdefault("username", user.get_username())
    except Exception:
        pass


def install_sql_hook() -> None:
    """If installed this causes Django's queries to be captured."""
    try:
        from django.db.backends.utils import CursorWrapper
    except ImportError:
        from django.db.backends.util import CursorWrapper

    try:
        # django 1.6 and 1.7 compatability
        from django.db.backends import BaseDatabaseWrapper
    except ImportError:
        # django 1.8 or later
        from django.db.backends.base.base import BaseDatabaseWrapper

    try:
        real_execute = CursorWrapper.execute
        real_executemany = CursorWrapper.executemany
        real_connect = BaseDatabaseWrapper.connect
        real_commit = BaseDatabaseWrapper._commit
        real_rollback = BaseDatabaseWrapper._rollback
    except AttributeError:
        # This won't work on Django versions < 1.6
        return

    @ensure_integration_enabled(DjangoIntegration, real_execute)
    def execute(
        self: "CursorWrapper", sql: "Any", params: "Optional[Any]" = None
    ) -> "Any":
        with record_sql_queries(
            cursor=self.cursor,
            query=sql,
            params_list=params,
            paramstyle="format",
            executemany=False,
            span_origin=DjangoIntegration.origin_db,
        ) as span:
            _set_db_data(span, self)
            result = real_execute(self, sql, params)

            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return result

    @ensure_integration_enabled(DjangoIntegration, real_executemany)
    def executemany(
        self: "CursorWrapper", sql: "Any", param_list: "List[Any]"
    ) -> "Any":
        with record_sql_queries(
            cursor=self.cursor,
            query=sql,
            params_list=param_list,
            paramstyle="format",
            executemany=True,
            span_origin=DjangoIntegration.origin_db,
        ) as span:
            _set_db_data(span, self)

            result = real_executemany(self, sql, param_list)

            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return result

    @ensure_integration_enabled(DjangoIntegration, real_connect)
    def connect(self: "BaseDatabaseWrapper") -> None:
        with capture_internal_exceptions():
            sentry_sdk.add_breadcrumb(message="connect", category="query")

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return real_connect(self)
            with sentry_sdk.traces.start_span(
                name="connect",
                attributes={
                    "sentry.op": OP.DB,
                    "sentry.origin": DjangoIntegration.origin_db,
                },
            ) as span:
                _set_db_data(span, self)
                return real_connect(self)
        else:
            with sentry_sdk.start_span(
                op=OP.DB,
                name="connect",
                origin=DjangoIntegration.origin_db,
            ) as span:
                _set_db_data(span, self)
                return real_connect(self)

    def _commit(self: "BaseDatabaseWrapper") -> None:
        integration = sentry_sdk.get_client().get_integration(DjangoIntegration)

        if integration is None or not integration.db_transaction_spans:
            return real_commit(self)

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return real_commit(self)
            with sentry_sdk.traces.start_span(
                name=SPANNAME.DB_COMMIT,
                attributes={
                    "sentry.op": OP.DB,
                    "sentry.origin": DjangoIntegration.origin_db,
                },
            ) as span:
                _set_db_data(span, self, SPANNAME.DB_COMMIT)
                return real_commit(self)
        else:
            with sentry_sdk.start_span(
                op=OP.DB,
                name=SPANNAME.DB_COMMIT,
                origin=DjangoIntegration.origin_db,
            ) as span:
                _set_db_data(span, self, SPANNAME.DB_COMMIT)
                return real_commit(self)

    def _rollback(self: "BaseDatabaseWrapper") -> None:
        integration = sentry_sdk.get_client().get_integration(DjangoIntegration)

        if integration is None or not integration.db_transaction_spans:
            return real_rollback(self)

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return real_rollback(self)
            with sentry_sdk.traces.start_span(
                name=SPANNAME.DB_ROLLBACK,
                attributes={
                    "sentry.op": OP.DB,
                    "sentry.origin": DjangoIntegration.origin_db,
                },
            ) as span:
                _set_db_data(span, self, SPANNAME.DB_ROLLBACK)
                return real_rollback(self)
        else:
            with sentry_sdk.start_span(
                op=OP.DB,
                name=SPANNAME.DB_ROLLBACK,
                origin=DjangoIntegration.origin_db,
            ) as span:
                _set_db_data(span, self, SPANNAME.DB_ROLLBACK)
                return real_rollback(self)

    CursorWrapper.execute = execute
    CursorWrapper.executemany = executemany
    BaseDatabaseWrapper.connect = connect
    BaseDatabaseWrapper._commit = _commit
    BaseDatabaseWrapper._rollback = _rollback
    ignore_logger("django.db.backends")


def _set_db_data(
    span: "Union[Span, StreamedSpan]",
    cursor_or_db: "Any",
    db_operation: "Optional[str]" = None,
) -> None:
    db = cursor_or_db.db if hasattr(cursor_or_db, "db") else cursor_or_db
    vendor = db.vendor
    if isinstance(span, StreamedSpan):
        span.set_attribute(SPANDATA.DB_SYSTEM_NAME, vendor)

        if db_operation is not None:
            span.set_attribute(SPANDATA.DB_OPERATION_NAME, db_operation)
    else:
        span.set_data(SPANDATA.DB_SYSTEM, vendor)

        if db_operation is not None:
            span.set_data(SPANDATA.DB_OPERATION, db_operation)

    # Some custom backends override `__getattr__`, making it look like `cursor_or_db`
    # actually has a `connection` and the `connection` has a `get_dsn_parameters`
    # attribute, only to throw an error once you actually want to call it.
    # Hence the `inspect` check whether `get_dsn_parameters` is an actual callable
    # function.
    is_psycopg2 = (
        hasattr(cursor_or_db, "connection")
        and hasattr(cursor_or_db.connection, "get_dsn_parameters")
        and inspect.isroutine(cursor_or_db.connection.get_dsn_parameters)
    )
    if is_psycopg2:
        connection_params = cursor_or_db.connection.get_dsn_parameters()
    else:
        try:
            # psycopg3, only extract needed params as get_parameters
            # can be slow because of the additional logic to filter out default
            # values
            connection_params = {
                "dbname": cursor_or_db.connection.info.dbname,
                "port": cursor_or_db.connection.info.port,
            }
            # PGhost returns host or base dir of UNIX socket as an absolute path
            # starting with /, use it only when it contains host
            pg_host = cursor_or_db.connection.info.host
            if pg_host and not pg_host.startswith("/"):
                connection_params["host"] = pg_host
        except Exception:
            connection_params = db.get_connection_params()

    db_name = connection_params.get("dbname") or connection_params.get("database")

    if isinstance(span, StreamedSpan):
        if db_name is not None:
            span.set_attribute(SPANDATA.DB_NAMESPACE, db_name)

        set_on_span = span.set_attribute
    else:
        if db_name is not None:
            span.set_data(SPANDATA.DB_NAME, db_name)

        set_on_span = span.set_data

    server_address = connection_params.get("host")
    if server_address is not None:
        set_on_span(SPANDATA.SERVER_ADDRESS, server_address)

    server_port = connection_params.get("port")
    if server_port is not None:
        set_on_span(SPANDATA.SERVER_PORT, str(server_port))

    server_socket_address = connection_params.get("unix_socket")
    if server_socket_address is not None:
        set_on_span(SPANDATA.SERVER_SOCKET_ADDRESS, server_socket_address)


def add_template_context_repr_sequence() -> None:
    try:
        from django.template.context import BaseContext

        add_repr_sequence_type(BaseContext)
    except Exception:
        pass



from contextlib import contextmanager
from typing import Any, Generator

import sentry_sdk
from sentry_sdk import start_span
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import BAGGAGE_HEADER_NAME
from sentry_sdk.tracing_utils import (
    add_http_breadcrumb,
    add_http_request_source,
    add_sentry_baggage_to_headers,
    has_span_streaming_enabled,
    propagate_trace_headers,
    should_propagate_trace,
)
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    capture_internal_exceptions,
    logger,
    parse_url,
)

try:
    from pyreqwest.client import (  # type: ignore[import-not-found]
        ClientBuilder,
        SyncClientBuilder,
    )
    from pyreqwest.middleware import Next, SyncNext  # type: ignore[import-not-found]
    from pyreqwest.request import (  # type: ignore[import-not-found]
        OneOffRequestBuilder,
        Request,
        SyncOneOffRequestBuilder,
    )
    from pyreqwest.response import (  # type: ignore[import-not-found]
        Response,
        SyncResponse,
    )
except ImportError:
    raise DidNotEnable("pyreqwest not installed or incompatible version installed")


class PyreqwestIntegration(Integration):
    identifier = "pyreqwest"
    origin = f"auto.http.{identifier}"

    @staticmethod
    def setup_once() -> None:
        _patch_pyreqwest()


def _patch_pyreqwest() -> None:
    # Patch Client Builders
    _patch_builder_method(ClientBuilder, "build", sentry_async_middleware)
    _patch_builder_method(SyncClientBuilder, "build", sentry_sync_middleware)

    # Patch Request Builders
    _patch_builder_method(OneOffRequestBuilder, "send", sentry_async_middleware)
    _patch_builder_method(SyncOneOffRequestBuilder, "send", sentry_sync_middleware)


def _patch_builder_method(cls: type, method_name: str, middleware: "Any") -> None:
    if not hasattr(cls, method_name):
        return

    original_method = getattr(cls, method_name)

    def sentry_patched_method(self: "Any", *args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(PyreqwestIntegration)

        if getattr(self, "_sentry_instrumented", False) or integration is None:
            return original_method(self, *args, **kwargs)

        self.with_middleware(middleware)

        try:
            self._sentry_instrumented = True
        except (TypeError, AttributeError):
            # In case the instance itself is immutable or doesn't allow extra attributes
            pass

        return original_method(self, *args, **kwargs)

    setattr(cls, method_name, sentry_patched_method)


@contextmanager
def _sentry_pyreqwest_span(request: "Request") -> "Generator[Any, None, None]":
    parsed_url = None
    with capture_internal_exceptions():
        parsed_url = parse_url(str(request.url), sanitize=False)

    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        if sentry_sdk.traces.get_current_span() is None:
            propagate_trace_headers(client=sentry_sdk.get_client(), request=request)
            yield None
            return

        with sentry_sdk.traces.start_span(
            name=f"{request.method} {parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE}",
            attributes={
                "sentry.op": OP.HTTP_CLIENT,
                "sentry.origin": PyreqwestIntegration.origin,
                SPANDATA.HTTP_REQUEST_METHOD: request.method,
            },
        ) as span:
            if parsed_url is not None and should_send_default_pii():
                span.set_attribute(SPANDATA.URL_FULL, parsed_url.url)
                span.set_attribute(SPANDATA.URL_QUERY, parsed_url.query)
                span.set_attribute(SPANDATA.URL_FRAGMENT, parsed_url.fragment)

            propagate_trace_headers(client=sentry_sdk.get_client(), request=request)

            yield span

            if span is not None:
                with capture_internal_exceptions():
                    add_http_request_source(span)

            return

    with start_span(
        op=OP.HTTP_CLIENT,
        name=f"{request.method} {parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE}",
        origin=PyreqwestIntegration.origin,
    ) as span:
        span.set_data(SPANDATA.HTTP_METHOD, request.method)
        if parsed_url is not None:
            span.set_data("url", parsed_url.url)
            span.set_data(SPANDATA.HTTP_QUERY, parsed_url.query)
            span.set_data(SPANDATA.HTTP_FRAGMENT, parsed_url.fragment)

        if should_propagate_trace(sentry_sdk.get_client(), str(request.url)):
            for (
                key,
                value,
            ) in sentry_sdk.get_current_scope().iter_trace_propagation_headers():
                logger.debug(
                    "[Tracing] Adding `{key}` header {value} to outgoing request to {url}.".format(
                        key=key, value=value, url=request.url
                    )
                )

                if key == BAGGAGE_HEADER_NAME:
                    add_sentry_baggage_to_headers(request.headers, value)
                else:
                    request.headers[key] = value

        yield span

    with capture_internal_exceptions():
        add_http_request_source(span)


async def sentry_async_middleware(
    request: "Request",
    next_handler: "Next",
) -> "Response":
    if sentry_sdk.get_client().get_integration(PyreqwestIntegration) is None:
        return await next_handler.run(request)

    method = request.method
    parsed_url = None
    with capture_internal_exceptions():
        # This needs to be done early because the URL is no longer accessible
        # after the request has been sent
        parsed_url = parse_url(str(request.url), sanitize=False)

    response = None
    with _sentry_pyreqwest_span(request) as span:
        response = await next_handler.run(request)
        if isinstance(span, StreamedSpan):
            span.status = "error" if response.status >= 400 else "ok"
            span.set_attribute(
                SPANDATA.HTTP_STATUS_CODE,
                response.status,
            )
        elif span is not None:
            span.set_http_status(response.status)

    if response is not None:
        breadcrumb_data = {
            SPANDATA.HTTP_METHOD: method,
            SPANDATA.HTTP_STATUS_CODE: response.status,
        }

        if parsed_url and should_send_default_pii():
            breadcrumb_data.update(
                {
                    "url": parsed_url.url,
                    SPANDATA.HTTP_QUERY: parsed_url.query,
                    SPANDATA.HTTP_FRAGMENT: parsed_url.fragment,
                }
            )

        add_http_breadcrumb(response.status, breadcrumb_data)

    return response


def sentry_sync_middleware(
    request: "Request", next_handler: "SyncNext"
) -> "SyncResponse":
    if sentry_sdk.get_client().get_integration(PyreqwestIntegration) is None:
        return next_handler.run(request)

    method = request.method
    parsed_url = None
    with capture_internal_exceptions():
        # This needs to be done early because the URL is no longer accessible
        # after the request has been sent
        parsed_url = parse_url(str(request.url), sanitize=False)

    response = None
    with _sentry_pyreqwest_span(request) as span:
        response = next_handler.run(request)
        if isinstance(span, StreamedSpan):
            span.status = "error" if response.status >= 400 else "ok"
            span.set_attribute(
                SPANDATA.HTTP_STATUS_CODE,
                response.status,
            )
        elif span is not None:
            span.set_http_status(response.status)

    if response is not None:
        breadcrumb_data = {
            SPANDATA.HTTP_METHOD: method,
            SPANDATA.HTTP_STATUS_CODE: response.status,
        }

        if parsed_url and should_send_default_pii():
            breadcrumb_data.update(
                {
                    "url": parsed_url.url,
                    SPANDATA.HTTP_QUERY: parsed_url.query,
                    SPANDATA.HTTP_FRAGMENT: parsed_url.fragment,
                }
            )

        add_http_breadcrumb(response.status, breadcrumb_data)

    return response



from typing import TYPE_CHECKING, Any, List, Optional, TypedDict, Union

import sentry_sdk
from sentry_sdk.ai.utils import set_data_normalized
from sentry_sdk.consts import SPANDATA
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.utils import (
    has_data_collection_enabled,
    safe_serialize,
)

from .utils import (
    UsageData,
    extract_contents_text,
    extract_finish_reasons,
    extract_tool_calls,
    extract_usage_data,
)

if TYPE_CHECKING:
    from google.genai.types import GenerateContentResponse

    from sentry_sdk.tracing import Span


class AccumulatedResponse(TypedDict):
    id: "Optional[str]"
    model: "Optional[str]"
    text: str
    finish_reasons: "List[str]"
    tool_calls: "List[dict[str, Any]]"
    usage_metadata: "Optional[UsageData]"


def element_wise_usage_max(self: "UsageData", other: "UsageData") -> "UsageData":
    return UsageData(
        input_tokens=max(self["input_tokens"], other["input_tokens"]),
        output_tokens=max(self["output_tokens"], other["output_tokens"]),
        input_tokens_cached=max(
            self["input_tokens_cached"], other["input_tokens_cached"]
        ),
        output_tokens_reasoning=max(
            self["output_tokens_reasoning"], other["output_tokens_reasoning"]
        ),
        total_tokens=max(self["total_tokens"], other["total_tokens"]),
    )


def accumulate_streaming_response(
    chunks: "List[GenerateContentResponse]",
) -> "AccumulatedResponse":
    """Accumulate streaming chunks into a single response-like object."""
    accumulated_text = []
    finish_reasons = []
    tool_calls = []
    usage_data = None
    response_id = None
    model = None

    for chunk in chunks:
        # Extract text and tool calls
        if getattr(chunk, "candidates", None):
            for candidate in getattr(chunk, "candidates", []):
                if hasattr(candidate, "content") and getattr(
                    candidate.content, "parts", []
                ):
                    extracted_text = extract_contents_text(candidate.content)
                    if extracted_text:
                        accumulated_text.append(extracted_text)

        extracted_finish_reasons = extract_finish_reasons(chunk)
        if extracted_finish_reasons:
            finish_reasons.extend(extracted_finish_reasons)

        extracted_tool_calls = extract_tool_calls(chunk)
        if extracted_tool_calls:
            tool_calls.extend(extracted_tool_calls)

        # Use last possible chunk, in case of interruption, and
        # gracefully handle missing intermediate tokens by taking maximum
        # with previous token reporting.
        chunk_usage_data = extract_usage_data(chunk)
        usage_data = (
            chunk_usage_data
            if usage_data is None
            else element_wise_usage_max(usage_data, chunk_usage_data)
        )

    accumulated_response = AccumulatedResponse(
        text="".join(accumulated_text),
        finish_reasons=finish_reasons,
        tool_calls=tool_calls,
        usage_metadata=usage_data,
        id=response_id,
        model=model,
    )

    return accumulated_response


def set_span_data_for_streaming_response(
    span: "Union[Span, StreamedSpan]",
    integration: "Any",
    accumulated_response: "AccumulatedResponse",
) -> None:
    """Set span data for accumulated streaming response."""
    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    client = sentry_sdk.get_client()

    if accumulated_response.get("finish_reasons"):
        set_data_normalized(
            span,
            SPANDATA.GEN_AI_RESPONSE_FINISH_REASONS,
            accumulated_response["finish_reasons"],
        )

    response_id = accumulated_response.get("id")
    if response_id is not None:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_ID, response_id)

    response_model = accumulated_response.get("model")
    if response_model is not None:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, response_model)

    if accumulated_response["usage_metadata"] is None:
        return

    if accumulated_response["usage_metadata"]["input_tokens"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_INPUT_TOKENS,
            accumulated_response["usage_metadata"]["input_tokens"],
        )

    if accumulated_response["usage_metadata"]["input_tokens_cached"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_INPUT_TOKENS_CACHED,
            accumulated_response["usage_metadata"]["input_tokens_cached"],
        )

    if accumulated_response["usage_metadata"]["output_tokens"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS,
            accumulated_response["usage_metadata"]["output_tokens"],
        )

    if accumulated_response["usage_metadata"]["output_tokens_reasoning"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS_REASONING,
            accumulated_response["usage_metadata"]["output_tokens_reasoning"],
        )

    if accumulated_response["usage_metadata"]["total_tokens"]:
        set_on_span(
            SPANDATA.GEN_AI_USAGE_TOTAL_TOKENS,
            accumulated_response["usage_metadata"]["total_tokens"],
        )

    if accumulated_response.get("tool_calls"):
        if has_data_collection_enabled(client.options):
            if client.options["data_collection"]["gen_ai"]["outputs"]:
                set_on_span(
                    SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS,
                    safe_serialize(accumulated_response["tool_calls"]),
                )
        else:
            # Before data collection was introduced this was unconditionally set
            set_on_span(
                SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS,
                safe_serialize(accumulated_response["tool_calls"]),
            )

    if accumulated_response.get("text"):
        if has_data_collection_enabled(client.options):
            if client.options["data_collection"]["gen_ai"]["outputs"]:
                set_on_span(
                    SPANDATA.GEN_AI_RESPONSE_TEXT,
                    safe_serialize([accumulated_response["text"]]),
                )

        elif should_send_default_pii() and integration.include_prompts:
            set_on_span(
                SPANDATA.GEN_AI_RESPONSE_TEXT,
                safe_serialize([accumulated_response["text"]]),
            )



try:
    from re import Pattern
except ImportError:
    # 3.6
    from typing import Pattern

from typing import TYPE_CHECKING, TypeVar, Union

# Re-exported for compat, since code out there in the wild might use this variable.
MYPY = TYPE_CHECKING


SENSITIVE_DATA_SUBSTITUTE = "[Filtered]"
BLOB_DATA_SUBSTITUTE = "[Blob substitute]"
OVER_SIZE_LIMIT_SUBSTITUTE = "[Exceeds maximum size]"
UNPARSABLE_RAW_DATA_SUBSTITUTE = "[Unparsable]"


class AnnotatedValue:
    """
    Meta information for a data field in the event payload.
    This is to tell Relay that we have tampered with the fields value.
    See:
    https://github.com/getsentry/relay/blob/be12cd49a0f06ea932ed9b9f93a655de5d6ad6d1/relay-general/src/types/meta.rs#L407-L423
    """

    __slots__ = ("value", "metadata")

    def __init__(self, value: "Optional[Any]", metadata: "Dict[str, Any]") -> None:
        self.value = value
        self.metadata = metadata

    def __eq__(self, other: "Any") -> bool:
        if not isinstance(other, AnnotatedValue):
            return False

        return self.value == other.value and self.metadata == other.metadata

    def __str__(self: "AnnotatedValue") -> str:
        return str({"value": str(self.value), "metadata": str(self.metadata)})

    def __len__(self: "AnnotatedValue") -> int:
        if self.value is not None:
            return len(self.value)
        else:
            return 0

    @classmethod
    def removed_because_raw_data(cls) -> "AnnotatedValue":
        """The value was removed because it could not be parsed. This is done for request body values that are not json nor a form."""
        # This is the legacy approach - we want to transition over to `substituted_because_raw_data` after we completely transition
        # to span-first
        return AnnotatedValue(
            value="",
            metadata={
                "rem": [  # Remark
                    [
                        "!raw",  # Unparsable raw data
                        "x",  # The fields original value was removed
                    ]
                ]
            },
        )

    @classmethod
    def substituted_because_raw_data(cls) -> "AnnotatedValue":
        """The value was replaced because it could not be parsed. This is done for request body values that are not json nor a form."""
        return AnnotatedValue(
            value=UNPARSABLE_RAW_DATA_SUBSTITUTE,
            metadata={
                "rem": [  # Remark
                    [
                        "!raw",  # Unparsable raw data
                        "s",  # The fields original value was substituted
                    ]
                ]
            },
        )

    @classmethod
    def removed_because_over_size_limit(cls, value: "Any" = "") -> "AnnotatedValue":
        """
        The actual value was removed because the size of the field exceeded the configured maximum size,
        for example specified with the max_request_body_size sdk option.
        """
        # This is the legacy approach - we want to transition over to `substituted_because_over_size_limit` after we completely transition
        # to span-first
        return AnnotatedValue(
            value=value,
            metadata={
                "rem": [  # Remark
                    [
                        "!config",  # Because of configured maximum size
                        "x",  # The fields original value was removed
                    ]
                ]
            },
        )

    @classmethod
    def substituted_because_over_size_limit(
        cls, value: "Any" = OVER_SIZE_LIMIT_SUBSTITUTE
    ) -> "AnnotatedValue":
        """
        The actual value was replaced because the size of the field exceeded the configured maximum size,
        for example specified with the max_request_body_size sdk option.
        """
        return AnnotatedValue(
            value=value,
            metadata={
                "rem": [  # Remark
                    [
                        "!config",  # Because of configured maximum size
                        "s",  # The fields original value was substituted
                    ]
                ]
            },
        )

    @classmethod
    def substituted_because_contains_sensitive_data(cls) -> "AnnotatedValue":
        """The actual value was removed because it contained sensitive information."""
        return AnnotatedValue(
            value=SENSITIVE_DATA_SUBSTITUTE,
            metadata={
                "rem": [  # Remark
                    [
                        "!config",  # Because of SDK configuration (in this case the config is the hard coded removal of certain django cookies)
                        "s",  # The fields original value was substituted
                    ]
                ]
            },
        )


T = TypeVar("T")
Annotated = Union[AnnotatedValue, T]


if TYPE_CHECKING:
    from collections.abc import Container, MutableMapping, Sequence
    from datetime import datetime
    from types import TracebackType
    from typing import Any, Callable, Dict, List, Mapping, NotRequired, Optional, Type

    from typing_extensions import Literal, TypedDict

    import sentry_sdk

    class SDKInfo(TypedDict):
        name: str
        version: str
        packages: "Sequence[Mapping[str, str]]"

    class KeyValueCollectionBehaviour(TypedDict):
        mode: 'Literal["off", "denylist", "allowlist"]'
        terms: "NotRequired[List[str]]"

    class GenAICollectionUserOptions(TypedDict, total=False):
        inputs: bool
        outputs: bool

    class GenAICollectionBehaviour(TypedDict):
        inputs: bool
        outputs: bool

    class GraphQLCollectionUserOptions(TypedDict, total=False):
        document: bool
        variables: bool

    class GraphQLCollectionBehaviour(TypedDict):
        document: bool
        variables: bool

    class HttpHeadersCollectionUserOptions(TypedDict, total=False):
        request: "KeyValueCollectionBehaviour"

    class HttpHeadersCollectionBehaviour(TypedDict):
        request: "KeyValueCollectionBehaviour"

    class DataCollectionUserOptions(TypedDict, total=False):
        user_info: bool
        cookies: "KeyValueCollectionBehaviour"
        http_headers: "HttpHeadersCollectionUserOptions"
        http_bodies: "List[str]"
        url_query_params: "KeyValueCollectionBehaviour"
        graphql: "GraphQLCollectionUserOptions"
        gen_ai: "GenAICollectionUserOptions"
        database_query_data: bool
        queues: bool
        stack_frame_variables: bool
        frame_context_lines: int

    class DataCollection(TypedDict):
        provided_by_user: bool
        user_info: bool
        cookies: "KeyValueCollectionBehaviour"
        http_headers: "HttpHeadersCollectionBehaviour"
        http_bodies: "List[str]"
        url_query_params: "KeyValueCollectionBehaviour"
        graphql: "GraphQLCollectionBehaviour"
        gen_ai: "GenAICollectionBehaviour"
        database_query_data: bool
        queues: bool
        stack_frame_variables: bool
        frame_context_lines: int

    # "critical" is an alias of "fatal" recognized by Relay
    LogLevelStr = Literal["fatal", "critical", "error", "warning", "info", "debug"]

    DurationUnit = Literal[
        "nanosecond",
        "microsecond",
        "millisecond",
        "second",
        "minute",
        "hour",
        "day",
        "week",
    ]

    InformationUnit = Literal[
        "bit",
        "byte",
        "kilobyte",
        "kibibyte",
        "megabyte",
        "mebibyte",
        "gigabyte",
        "gibibyte",
        "terabyte",
        "tebibyte",
        "petabyte",
        "pebibyte",
        "exabyte",
        "exbibyte",
    ]

    FractionUnit = Literal["ratio", "percent"]
    MeasurementUnit = Union[DurationUnit, InformationUnit, FractionUnit, str]

    MeasurementValue = TypedDict(
        "MeasurementValue",
        {
            "value": float,
            "unit": NotRequired[Optional[MeasurementUnit]],
        },
    )

    Event = TypedDict(
        "Event",
        {
            "breadcrumbs": Annotated[
                dict[Literal["values"], list[dict[str, Any]]]
            ],  # TODO: We can expand on this type
            "check_in_id": str,
            "contexts": dict[str, dict[str, object]],
            "dist": str,
            "duration": Optional[float],
            "environment": Optional[str],
            "errors": list[dict[str, Any]],  # TODO: We can expand on this type
            "event_id": str,
            "exception": dict[
                Literal["values"], list[dict[str, Any]]
            ],  # TODO: We can expand on this type
            "extra": MutableMapping[str, object],
            "fingerprint": list[str],
            "level": LogLevelStr,
            "logentry": Mapping[str, object],
            "logger": str,
            "measurements": dict[str, MeasurementValue],
            "message": str,
            "modules": dict[str, str],
            "monitor_config": Mapping[str, object],
            "monitor_slug": Optional[str],
            "platform": Literal["python"],
            "profile": object,  # Should be sentry_sdk.profiler.Profile, but we can't import that here due to circular imports
            "release": Optional[str],
            "request": dict[str, object],
            "sdk": Mapping[str, object],
            "server_name": str,
            "spans": Annotated[list[dict[str, object]]],
            "stacktrace": dict[
                str, object
            ],  # We access this key in the code, but I am unsure whether we ever set it
            "start_timestamp": datetime,
            "status": Optional[str],
            "tags": MutableMapping[
                str, str
            ],  # Tags must be less than 200 characters each
            "threads": dict[
                Literal["values"], list[dict[str, Any]]
            ],  # TODO: We can expand on this type
            "timestamp": Optional[datetime],  # Must be set before sending the event
            "transaction": str,
            "transaction_info": Mapping[str, Any],  # TODO: We can expand on this type
            "type": Literal["check_in", "transaction"],
            "user": dict[str, object],
            "_dropped_spans": int,
            "_has_gen_ai_span": bool,
        },
        total=False,
    )

    ExcInfo = Union[
        tuple[Type[BaseException], BaseException, Optional[TracebackType]],
        tuple[None, None, None],
    ]

    # TODO: Make a proper type definition for this (PRs welcome!)
    Hint = Dict[str, Any]

    AttributeValue = (
        str
        | bool
        | float
        | int
        | list[str]
        | list[bool]
        | list[float]
        | list[int]
        | tuple[str, ...]
        | tuple[bool, ...]
        | tuple[float, ...]
        | tuple[int, ...]
    )
    Attributes = dict[str, AttributeValue]

    SerializedAttributeValue = TypedDict(
        # https://develop.sentry.dev/sdk/telemetry/attributes/#supported-types
        "SerializedAttributeValue",
        {
            "type": Literal[
                "string",
                "boolean",
                "double",
                "integer",
                "array",
            ],
            "value": AttributeValue,
        },
    )

    Log = TypedDict(
        "Log",
        {
            "severity_text": str,
            "severity_number": int,
            "body": str,
            "attributes": Attributes,
            "time_unix_nano": int,
            "trace_id": Optional[str],
            "span_id": Optional[str],
        },
    )

    MetricType = Literal["counter", "gauge", "distribution"]
    MetricUnit = Union[DurationUnit, InformationUnit, str]

    Metric = TypedDict(
        "Metric",
        {
            "timestamp": float,
            "trace_id": Optional[str],
            "span_id": Optional[str],
            "name": str,
            "type": MetricType,
            "value": float,
            "unit": Optional[MetricUnit],
            "attributes": Attributes,
        },
    )

    MetricProcessor = Callable[[Metric, Hint], Optional[Metric]]

    SpanJSON = TypedDict(
        "SpanJSON",
        {
            "trace_id": str,
            "span_id": str,
            "parent_span_id": NotRequired[str],
            "name": str,
            "status": str,
            "is_segment": bool,
            "start_timestamp": float,
            "end_timestamp": NotRequired[float],
            "attributes": NotRequired[Attributes],
            "_segment_span": NotRequired["sentry_sdk.traces.StreamedSpan"],
        },
    )

    # TODO: Make a proper type definition for this (PRs welcome!)
    Breadcrumb = Dict[str, Any]

    # TODO: Make a proper type definition for this (PRs welcome!)
    BreadcrumbHint = Dict[str, Any]

    # TODO: Make a proper type definition for this (PRs welcome!)
    SamplingContext = Dict[str, Any]

    EventProcessor = Callable[[Event, Hint], Optional[Event]]
    ErrorProcessor = Callable[[Event, ExcInfo], Optional[Event]]
    BreadcrumbProcessor = Callable[[Breadcrumb, BreadcrumbHint], Optional[Breadcrumb]]
    TransactionProcessor = Callable[[Event, Hint], Optional[Event]]
    LogProcessor = Callable[[Log, Hint], Optional[Log]]

    TracesSampler = Callable[[SamplingContext], Union[float, int, bool]]

    # https://github.com/python/mypy/issues/5710
    NotImplementedType = Any

    EventDataCategory = Literal[
        "default",
        "error",
        "crash",
        "transaction",
        "security",
        "attachment",
        "session",
        "internal",
        "profile",
        "profile_chunk",
        "monitor",
        "span",
        "log_item",
        "log_byte",
        "trace_metric",
    ]
    SessionStatus = Literal["ok", "exited", "crashed", "abnormal"]

    ContinuousProfilerMode = Literal["thread", "gevent", "unknown"]
    ProfilerMode = Union[ContinuousProfilerMode, Literal["sleep"]]

    MonitorConfigScheduleType = Literal["crontab", "interval"]
    MonitorConfigScheduleUnit = Literal[
        "year",
        "month",
        "week",
        "day",
        "hour",
        "minute",
        "second",  # not supported in Sentry and will result in a warning
    ]

    MonitorConfigSchedule = TypedDict(
        "MonitorConfigSchedule",
        {
            "type": MonitorConfigScheduleType,
            "value": Union[int, str],
            "unit": MonitorConfigScheduleUnit,
        },
        total=False,
    )

    MonitorConfig = TypedDict(
        "MonitorConfig",
        {
            "schedule": MonitorConfigSchedule,
            "timezone": str,
            "checkin_margin": int,
            "max_runtime": int,
            "failure_issue_threshold": int,
            "recovery_threshold": int,
            "owner": str,
        },
        total=False,
    )

    HttpStatusCodeRange = Union[int, Container[int]]

    class TextPart(TypedDict):
        type: Literal["text"]
        content: str

    class ReasoningPart(TypedDict):
        type: Literal["reasoning"]
        content: str

    class ToolCallPart(TypedDict):
        type: Literal["tool_call"]
        name: NotRequired[str]
        arguments: NotRequired[Any]

    class ToolDefinition(TypedDict):
        type: str
        name: NotRequired[str]
        description: NotRequired[str]
        parameters: NotRequired[dict[str, object]]

    IgnoreSpansName = Union[str, Pattern[str]]
    IgnoreSpansContext = TypedDict(
        "IgnoreSpansContext",
        {"name": IgnoreSpansName, "attributes": Attributes},
        total=False,
    )
    IgnoreSpansConfig = list[Union[IgnoreSpansName, IgnoreSpansContext]]
